import ast
from typing import Dict, List, Tuple

from boa3.model.builtin.method.intmethod import IntMethod
from boa3.model.variable import Variable
from boa3.neo.vm.opcode import OpcodeHelper
from boa3.neo.vm.opcode.Opcode import Opcode


class IntByteStringMethod(IntMethod):

    def __init__(self):
        from boa3.model.type.type import Type

        args: Dict[str, Variable] = {
            'value': Variable(Type.union.build([
                Type.bytes,
                Type.str
            ])),
            'base': Variable(Type.int)
        }

        value_default = ast.parse("{0}".format(Type.int.default_value)
                                  ).body[0].value

        base_default = ast.parse("{0}".format(10)
                                 ).body[0].value

        super().__init__(args, [value_default, base_default])

    @property
    def _opcode(self) -> List[Tuple[Opcode, bytes]]:
        from boa3.compiler.codegenerator import get_bytes_count
        from boa3.neo.vm.type.StackItem import StackItemType
        jmp_place_holder = (Opcode.JMP, b'\x01')
        a_lower = ord('a')
        a_upper = ord('A')
        zero = ord('0')

        # region verify_base

        verify_base_ge37 = [    # verifies if base >= 37, base greater than 36 shouldn't be accepted
            (Opcode.OVER, b''),
            OpcodeHelper.get_push_and_data(37),
            jmp_place_holder,   # jumps to an assertion error
        ]

        verify_base_equal1 = [  # verifies if base == 1, base equals 1 shouldn't be accepted
            (Opcode.OVER, b''),
            (Opcode.PUSH1, b''),
            jmp_place_holder    # jumps to an assertion error
        ]

        verify_base_le_minus1 = [   # verifies if base <= -1, base less than 0 shouldn't be accepted
            (Opcode.OVER, b''),
            (Opcode.PUSHM1, b''),
            jmp_place_holder,       # jumps to an assertion error
        ]

        verify_base_jump_error = [  # jumps the next set of instructions
            jmp_place_holder
        ]

        verify_base_error = [       # an invalid base was used
            (Opcode.PUSH0, b''),
            (Opcode.ASSERT, b''),
        ]

        # region verify_base jump logic

        jump_instructions = verify_base_error
        verify_base_jump_error[-1] = OpcodeHelper.get_jump_and_data(Opcode.JMP, get_bytes_count(jump_instructions), True)

        jump_instructions = verify_base_jump_error
        verify_base_le_minus1[-1] = OpcodeHelper.get_jump_and_data(Opcode.JMPLE, get_bytes_count(jump_instructions), True)

        jump_instructions = verify_base_jump_error + verify_base_le_minus1
        verify_base_equal1[-1] = OpcodeHelper.get_jump_and_data(Opcode.JMPEQ, get_bytes_count(jump_instructions), True)

        jump_instructions = verify_base_jump_error + verify_base_le_minus1 + verify_base_equal1
        verify_base_ge37[-1] = OpcodeHelper.get_jump_and_data(Opcode.JMPGE, get_bytes_count(jump_instructions), True)

        # endregion

        verify_base = (
            verify_base_ge37 +
            verify_base_equal1 +
            verify_base_le_minus1 +
            verify_base_jump_error +
            verify_base_error
        )

        # endregion

        # region verify_code_literal

        verify_code_literal_first_char_is_0 = [     # verifies if the first char in the value is '0'
            (Opcode.DUP, b''),
            (Opcode.PUSH0, b''),
            (Opcode.PUSH1, b''),
            (Opcode.SUBSTR, b''),
            (Opcode.CONVERT, StackItemType.ByteString),
            OpcodeHelper.get_pushdata_and_data('0'),
            jmp_place_holder,                       # if first char is not 0, skip all verify_code_literal methods
        ]

        verify_code_literal_remove_first_char = [   # the first char is 0, so it will be removed from the value
            (Opcode.PUSH1, b''),
            (Opcode.OVER, b''),
            (Opcode.SIZE, b''),
            (Opcode.DEC, b''),
            (Opcode.SUBSTR, b''),
            (Opcode.CONVERT, StackItemType.ByteString),
        ]

        verify_code_literal_get_base_char = [       # puts the char that could refer to a base in the top of the stack
            (Opcode.DUP, b''),
            (Opcode.PUSH0, b''),
            (Opcode.PUSH1, b''),
            (Opcode.SUBSTR, b''),
            (Opcode.CONVERT, StackItemType.ByteString),
        ]

        verify_code_literal_initialize_verification = [     # puts a False in the top of the stack to assist the
            (Opcode.PUSH0, b'')                             # verifications above
        ]

        verify_code_literal_base_verification = (           # verifies if the original value starts with 0b, 0B, 0o, 0O
            self._verify_is_specific_base('b', 2) +         # 0x, or 0X, and changes the base if necessary
            self._verify_is_specific_base('B', 2) +
            self._verify_is_specific_base('o', 8) +
            self._verify_is_specific_base('O', 8) +
            self._verify_is_specific_base('x', 16) +
            self._verify_is_specific_base('X', 16)
        )

        verify_code_literal_drop_aux = [                    # drops auxiliary values from the stack and just keeps the
            (Opcode.DROP, b''),                             # base and value on it
            (Opcode.DROP, b''),
        ]

        # region verify_code_literal jump logic

        jump_instructions = (
            verify_code_literal_remove_first_char +
            verify_code_literal_get_base_char +
            verify_code_literal_initialize_verification +
            verify_code_literal_base_verification +
            verify_code_literal_drop_aux
        )

        verify_code_literal_first_char_is_0[-1] = OpcodeHelper.get_jump_and_data(Opcode.JMPNE_L,
                                                                                 get_bytes_count(jump_instructions), True)

        # endregion

        verify_code_literal = (
            verify_code_literal_first_char_is_0 +
            verify_code_literal_remove_first_char +
            verify_code_literal_get_base_char +
            verify_code_literal_initialize_verification +
            verify_code_literal_base_verification +
            verify_code_literal_drop_aux
        )

        # endregion

        # region ascii to int

        ascii_to_int_initialize = [     # initialize auxiliary variables on the stack
            (Opcode.DUP, b''),
            (Opcode.SIZE, b''),
            (Opcode.DEC, b''),          # index = len(value) - 1
            (Opcode.DUP, b''),
            (Opcode.PUSH0, b''),
            (Opcode.GE, b''),
            (Opcode.ASSERT, b''),       # verifies if value was empty
            (Opcode.PUSH0, b''),        # sum = 0
            (Opcode.PUSH1, b''),        # mult = 1
        ]

        ascii_to_int_while = [      # loops from left to right of the value to calculate the equivalent sum in base 10
            (Opcode.PUSH3, b''),
            (Opcode.PICK, b''),
            (Opcode.PUSH3, b''),
            (Opcode.PICK, b''),
            (Opcode.PUSH1, b''),
            (Opcode.SUBSTR, b''),   # char = value[index]
            (Opcode.CONVERT, StackItemType.Integer),
        ]

        ascii_to_int_verify_az_lower_lt = [     # verifies ord(char) is less than ord('a')
            (Opcode.DUP, b''),                  # if ord(char) is indeed less, go to verify with ord('A')
            OpcodeHelper.get_push_and_data(a_lower),
            jmp_place_holder                    # jump to ascii_to_int_verify_az_upper_lt
        ]

        ascii_to_int_verify_az_lower_gt = [     # verifies ord(char) is greater than base + 86
            (Opcode.DUP, b''),                  # the minimum base that accepts letters is base 11, that's why it's
            (Opcode.PUSH6, b''),                # adding 86
            (Opcode.PICK, b''),                 # if ord(char) is greater than base + 86, the char is invalid
            OpcodeHelper.get_push_and_data(86),
            (Opcode.ADD, b''),
            (Opcode.LE, b''),
            (Opcode.ASSERT, b''),               # if assert is False, then the char was indeed invalid
        ]

        ascii_to_int_verify_az_lower = [        # the char was in between 'a' and 'z',
            OpcodeHelper.get_push_and_data(87),       # so it will be converted to the respective base 10 number
            (Opcode.SUB, b''),
            (Opcode.OVER, b''),
            (Opcode.MUL, b''),
            (Opcode.PUSH2, b''),
            (Opcode.PICK, b''),
            (Opcode.ADD, b''),                  # sum = sum + (ord(char) - 55)) * mult
            (Opcode.REVERSE3, b''),
            (Opcode.DROP, b''),
            jmp_place_holder                    # jumps to ascii_to_int_verify_while
        ]

        ascii_to_int_verify_az_upper_lt = [     # verifies ord(char) is less than ord('A')
            (Opcode.DUP, b''),                  # if ord(char) is indeed less, go to verify with ord('0')
            OpcodeHelper.get_push_and_data(a_upper),
            jmp_place_holder                    # jump to ascii_to_int_verify_09_lt
        ]

        ascii_to_int_verify_az_upper_gt = [     # verifies ord(char) is greater than base + 54
            (Opcode.DUP, b''),                  # the minimum base that accepts letters is base 11, that's why it's
            (Opcode.PUSH6, b''),                # adding 54
            (Opcode.PICK, b''),                 # if ord(char) is greater than base + 54, the char is invalid
            OpcodeHelper.get_push_and_data(54),
            (Opcode.ADD, b''),
            (Opcode.LE, b''),
            (Opcode.ASSERT, b''),               # if assert is False, then the char was indeed invalid
        ]

        ascii_to_int_verify_az_upper = [        # the char was in between 'A' and 'Z',
            OpcodeHelper.get_push_and_data(55),       # so it will be converted to the respective base 10 number
            (Opcode.SUB, b''),
            (Opcode.OVER, b''),
            (Opcode.MUL, b''),
            (Opcode.PUSH2, b''),
            (Opcode.PICK, b''),
            (Opcode.ADD, b''),                  # sum = sum + (ord(char) - 55)) * mult
            (Opcode.REVERSE3, b''),
            (Opcode.DROP, b''),
            jmp_place_holder                    # jumps to ascii_to_int_verify_while
        ]

        ascii_to_int_verify_09_lt = [           # verifies ord(char) is less than ord('0')
            (Opcode.DUP, b''),                  # if ord(char) is indeed less, the char is invalid
            OpcodeHelper.get_push_and_data(zero),
            (Opcode.GE, b''),
            (Opcode.ASSERT, b''),               # if assert is False, then the char was indeed invalid
        ]

        ascii_to_int_verify_09_gt = [           # verifies ord(char) is greater than base + ord('0')
            (Opcode.DUP, b''),                  # if ord(char) is greater than base + ord('0'), the char is invalid
            (Opcode.PUSH6, b''),
            (Opcode.PICK, b''),
            OpcodeHelper.get_push_and_data(zero),
            (Opcode.ADD, b''),
            (Opcode.LE, b''),
            (Opcode.ASSERT, b''),               # if assert is False, then the char was indeed invalid
        ]

        ascii_to_int_verify_09 = [              # the char was in between '0' and '9',
            OpcodeHelper.get_push_and_data(zero),     # so it will be converted to the respective base 10 number
            (Opcode.SUB, b''),
            (Opcode.OVER, b''),
            (Opcode.MUL, b''),
            (Opcode.PUSH2, b''),
            (Opcode.PICK, b''),
            (Opcode.ADD, b''),                  # sum = sum + (ord(char) - ord('0')) * mult
            (Opcode.REVERSE3, b''),
            (Opcode.DROP, b''),
        ]

        ascii_to_int_verify_while = [           # verifies if the loop already went through all the chars in value
            (Opcode.REVERSE3, b''),
            (Opcode.DEC, b''),                  # index--
            (Opcode.REVERSE3, b''),
            (Opcode.PUSH4, b''),
            (Opcode.PICK, b''),
            (Opcode.MUL, b''),                  # mult = mult * base
            (Opcode.PUSH2, b''),
            (Opcode.PICK, b''),
            (Opcode.PUSH0, b''),
            # jmp back to ascii_to_int_while if index >= 0
        ]

        # region ascii to int jump logic

        jump_instructions = (
            ascii_to_int_verify_09_lt +
            ascii_to_int_verify_09_gt +
            ascii_to_int_verify_09
        )

        ascii_to_int_verify_az_upper[-1] = OpcodeHelper.get_jump_and_data(Opcode.JMP,
                                                                          get_bytes_count(jump_instructions), True)

        jump_instructions = (
            ascii_to_int_verify_az_upper_gt +
            ascii_to_int_verify_az_upper
        )

        ascii_to_int_verify_az_upper_lt[-1] = OpcodeHelper.get_jump_and_data(Opcode.JMPLT,
                                                                             get_bytes_count(jump_instructions), True)

        jump_instructions = (
            ascii_to_int_verify_az_upper_lt +
            ascii_to_int_verify_az_upper_gt +
            ascii_to_int_verify_az_upper +
            ascii_to_int_verify_09_lt +
            ascii_to_int_verify_09_gt +
            ascii_to_int_verify_09
        )

        ascii_to_int_verify_az_lower[-1] = OpcodeHelper.get_jump_and_data(Opcode.JMP,
                                                                          get_bytes_count(jump_instructions), True)

        jump_instructions = (
            ascii_to_int_verify_az_lower_gt +
            ascii_to_int_verify_az_lower
        )

        ascii_to_int_verify_az_lower_lt[-1] = OpcodeHelper.get_jump_and_data(Opcode.JMPLT,
                                                                             get_bytes_count(jump_instructions), True)

        jump_instructions = (
            ascii_to_int_while +
            ascii_to_int_verify_az_lower_lt +
            ascii_to_int_verify_az_lower_gt +
            ascii_to_int_verify_az_lower +
            ascii_to_int_verify_az_upper_lt +
            ascii_to_int_verify_az_upper_gt +
            ascii_to_int_verify_az_upper +
            ascii_to_int_verify_09_gt +
            ascii_to_int_verify_09_lt +
            ascii_to_int_verify_09 +
            ascii_to_int_verify_while
        )

        ascii_to_int_verify_while.append(OpcodeHelper.get_jump_and_data(Opcode.JMPGE,
                                                                        -get_bytes_count(jump_instructions), True))

        # endregion

        ascii_to_int = (
            ascii_to_int_initialize +
            ascii_to_int_while +
            ascii_to_int_verify_az_lower_lt +
            ascii_to_int_verify_az_lower_gt +
            ascii_to_int_verify_az_lower +
            ascii_to_int_verify_az_upper_lt +
            ascii_to_int_verify_az_upper_gt +
            ascii_to_int_verify_az_upper +
            ascii_to_int_verify_09_lt +
            ascii_to_int_verify_09_gt +
            ascii_to_int_verify_09 +
            ascii_to_int_verify_while
        )

        # endregion

        # region clean stack

        clean_stack = [         # clean everything but the number in base 10
            (Opcode.DROP, b''),
            (Opcode.NIP, b''),
            (Opcode.NIP, b''),
            (Opcode.NIP, b''),
        ]

        # endregion

        return (
            verify_base +
            verify_code_literal +
            ascii_to_int +
            clean_stack
        )

    def _verify_is_specific_base(self, char: str, base: int) -> List[Tuple[Opcode, bytes]]:
        from boa3.compiler.codegenerator import get_bytes_count
        from boa3.neo.vm.type.StackItem import StackItemType
        jmp_place_holder = (Opcode.JMP, b'\x01')

        verify_top_stack = [    # verify if True is in the top of the stack
            jmp_place_holder
        ]

        top_stack_is_true = [       # if True is indeed at the top, ignore the rest of this method
            (Opcode.PUSH1, b''),
            jmp_place_holder        # jumps everything below
        ]

        compare_char = [            # compares the char with b, B, o, O, x, or X
            (Opcode.DUP, b''),
            OpcodeHelper.get_pushdata_and_data(char),
            jmp_place_holder
        ]

        char_not_equal = [          # if char at the top of the stack is not b, B, o, O, x, or X,
            (Opcode.PUSH0, b''),    # put False at the top of the stack

            jmp_place_holder        # jumps everything below
        ]

        verify_base_is_the_same = [     # verify if the equivalent base is the same as current base
            (Opcode.PUSH2, b''),
            (Opcode.PICK, b''),
            OpcodeHelper.get_push_and_data(base),
            jmp_place_holder            # jump to remove_base_char if equivalent base is the same as current base
        ]

        verify_base_is_0 = [            # verify if the current base is 0
            (Opcode.PUSH2, b''),
            (Opcode.PICK, b''),
            (Opcode.PUSH0, b''),
            jmp_place_holder            # if base is not 0, jump to put_true_in_stack
        ]

        change_base = [                 # if base was 0, then change it to the equivalent base
            (Opcode.REVERSE3, b''),
            (Opcode.DROP, b''),
            OpcodeHelper.get_push_and_data(base),
            (Opcode.REVERSE3, b''),
        ]

        remove_base_char = [            # if base was 0 or 16, remove the char in value
            (Opcode.SWAP, b''),
            (Opcode.PUSH1, b''),
            (Opcode.OVER, b''),
            (Opcode.SIZE, b''),
            (Opcode.DEC, b''),
            (Opcode.SUBSTR, b''),
            (Opcode.CONVERT, StackItemType.ByteString),
            (Opcode.SWAP, b''),
        ]

        put_true_in_stack = [           # puts True in the top of the stack to skip the next verifications
            (Opcode.PUSH1, b'')
        ]

        # region jmp logic

        jump_instructions = change_base + remove_base_char
        verify_base_is_0[-1] = OpcodeHelper.get_jump_and_data(Opcode.JMPNE, get_bytes_count(jump_instructions), True)

        jump_instructions = change_base + verify_base_is_0
        verify_base_is_the_same[-1] = OpcodeHelper.get_jump_and_data(Opcode.JMPEQ, get_bytes_count(jump_instructions), True)

        jump_instructions = change_base + remove_base_char + put_true_in_stack + verify_base_is_0 + verify_base_is_the_same
        char_not_equal[-1] = OpcodeHelper.get_jump_and_data(Opcode.JMP, get_bytes_count(jump_instructions), True)

        jump_instructions = char_not_equal
        compare_char[-1] = OpcodeHelper.get_jump_and_data(Opcode.JMPEQ, get_bytes_count(jump_instructions), True)

        jump_instructions = (
            compare_char +
            char_not_equal +
            change_base +
            put_true_in_stack +
            verify_base_is_the_same +
            verify_base_is_0 +
            remove_base_char
        )
        top_stack_is_true[-1] = OpcodeHelper.get_jump_and_data(Opcode.JMP, get_bytes_count(jump_instructions), True)

        jump_instructions = top_stack_is_true
        verify_top_stack[-1] = OpcodeHelper.get_jump_and_data(Opcode.JMPIFNOT, get_bytes_count(jump_instructions), True)

        # endregion

        return (
            verify_top_stack +
            top_stack_is_true +
            compare_char +
            char_not_equal +
            verify_base_is_the_same +
            verify_base_is_0 +
            change_base +
            remove_base_char +
            put_true_in_stack
        )



from boa3.builtin.type import UInt160


def get_exec_fee_factor() -> int:
    """
    Gets the execution fee factor. This is a multiplier that can be adjusted by the committee to adjust the system fees
    for transactions.

    :return: the execution fee factor
    :rtype: int
    """
    pass


def get_fee_per_byte() -> int:
    """
    Gets the network fee per transaction byte.

    :return: the network fee per transaction byte
    :rtype: int
    """
    pass


def get_storage_price() -> int:
    """
    Gets the storage price.

    :return: the snapshot used to read data
    :rtype: int
    """
    pass


def is_blocked(account: UInt160) -> bool:
    """
    Determines whether the specified account is blocked.

    :param account: the account to be checked
    :type account: UInt160

    :return: whether the account is blocked or not
    :rtype: bool
    """
    pass



__all__ = ['FindOptionsType',
           'StorageContextType',
           'StorageDeleteMethod',
           'StorageFindMethod',
           'StorageGetContextMethod',
           'StorageGetReadOnlyContextMethod',
           'StorageGetMethod',
           'StorageMapType',
           'StoragePutMethod'
           ]

from boa3.model.builtin.interop.storage.findoptionstype import FindOptionsType
from boa3.model.builtin.interop.storage.storagecontext import StorageContextType
from boa3.model.builtin.interop.storage.storagedeletemethod import StorageDeleteMethod
from boa3.model.builtin.interop.storage.storagefindmethod import StorageFindMethod
from boa3.model.builtin.interop.storage.storagegetcontextmethod import StorageGetContextMethod
from boa3.model.builtin.interop.storage.storagegetmethod import StorageGetMethod
from boa3.model.builtin.interop.storage.storagegetreadonlycontextmethod import StorageGetReadOnlyContextMethod
from boa3.model.builtin.interop.storage.storagemap.storagemaptype import StorageMapType
from boa3.model.builtin.interop.storage.storageputmethod import StoragePutMethod



from __future__ import annotations

from abc import ABC, abstractmethod
from typing import Dict, List, Tuple

from boa3.compiler.codegenerator import get_bytes_count
from boa3.model.symbol import ISymbol
from boa3.model.type.itype import IType
from boa3.neo.vm.opcode import OpcodeHelper
from boa3.neo.vm.opcode.Opcode import Opcode
from boa3.neo.vm.type.AbiType import AbiType
from boa3.neo.vm.type.Integer import Integer
from boa3.neo.vm.type.StackItem import StackItemType


class ClassType(IType, ABC):
    """
    An abstract class used to represent Python class
    """

    def __init__(self, identifier: str, decorators: list = None,
                 bases: List[ClassType] = None):
        super().__init__(identifier)

        if decorators is None:
            decorators = []
        else:
            # avoid circular import
            from boa3.model.decorator import IDecorator
            decorators = [decorator for decorator in decorators
                          if isinstance(decorator, IDecorator)]

        self.decorators = decorators

        if not isinstance(bases, list):
            bases = []
        self.bases: List[ClassType] = bases

    @property
    @abstractmethod
    def class_variables(self):
        """
        :rtype: Dict[str, boa3.model.variable.Variable]
        """
        class_vars = {}
        for base in self.bases:
            class_vars.update(base.class_variables)
        return class_vars

    @property
    @abstractmethod
    def instance_variables(self):
        """
        :rtype: Dict[str, boa3.model.variable.Variable]
        """
        instance_vars = {}
        for base in self.bases:
            instance_vars.update(base.instance_variables)
        return instance_vars

    @property
    def variables(self):
        """
        :rtype: Dict[str, boa3.model.variable.Variable]
        """
        variables = self.class_variables.copy()
        variables.update(self.instance_variables)
        return variables

    @property
    def _all_variables(self):
        """
        :rtype: Dict[str, boa3.model.variable.Variable]
        """
        return self.variables

    @property
    @abstractmethod
    def properties(self):
        """
        :rtype: Dict[str, boa3.model.property.Property]
        """
        props = {}
        for base in self.bases:
            props.update(base.properties)
        return props

    @property
    @abstractmethod
    def static_methods(self):
        """
        :rtype: Dict[str, boa3.model.method.Method]
        """
        static_funcs = {}
        for base in self.bases:
            static_funcs.update(base.static_methods)
        return static_funcs

    @property
    @abstractmethod
    def class_methods(self):
        """
        :rtype: Dict[str, boa3.model.method.Method]
        """
        class_funcs = {}
        for base in self.bases:
            class_funcs.update(base.class_methods)
        return class_funcs

    @property
    @abstractmethod
    def instance_methods(self):
        """
        :rtype: Dict[str, boa3.model.method.Method]
        """
        instance_funcs = {}
        for base in self.bases:
            instance_funcs.update(base.instance_methods)
        return instance_funcs

    @abstractmethod
    def constructor_method(self):
        """
        If the class constructor is None, it mustn't allow instantiation of this class

        :rtype: boa3.model.method.Method or None
        """
        pass

    @property
    def symbols(self) -> Dict[str, ISymbol]:
        s = {}
        s.update(self.static_methods)
        s.update(self.class_methods)
        s.update(self.instance_methods)
        s.update(self.variables)
        s.update(self.properties)
        return s

    @property
    def class_symbols(self) -> Dict[str, ISymbol]:
        s: Dict[str, ISymbol] = {}
        s.update(self.class_methods)  # class methods and variables can be accessed both
        s.update(self.class_variables)  # from class name or instance object
        s.update(self.static_methods)
        return s

    @property
    def instance_symbols(self) -> Dict[str, ISymbol]:
        s: Dict[str, ISymbol] = {}
        s.update(self.class_methods)  # class methods and variables can be accessed both
        s.update(self.class_variables)  # from class name or instance object
        s.update(self.instance_methods)
        s.update(self.instance_variables)
        s.update(self.properties)
        return s

    @property
    @abstractmethod
    def abi_type(self) -> AbiType:
        # must be overwritten, classes cannot be mapped to Any
        return super().abi_type

    @property
    @abstractmethod
    def stack_item(self) -> StackItemType:
        # must be overwritten, classes cannot be mapped to Any
        return super().stack_item

    @property
    def is_interface(self) -> bool:
        # TODO: change when other interfaces identifiers are implemented
        from boa3.model.builtin.decorator import ContractDecorator
        return any(isinstance(decorator, ContractDecorator) for decorator in self.decorators)

    def is_instance_opcodes(self) -> List[Tuple[Opcode, bytes]]:
        is_type_opcodes = [
            (Opcode.DUP, b''),  # if is the same internal type
            (Opcode.ISTYPE, self.stack_item)
        ]

        return_false = [
            (Opcode.DROP, b''),
            (Opcode.PUSH0, b''),  # return False
        ]
        return_false_bytes_size = get_bytes_count(return_false)

        final_instructions = []
        final_jmp = [(Opcode.JMP, Integer(2 + return_false_bytes_size).to_byte_array(min_length=1))]
        validate_type = self._is_instance_inner_opcodes(get_bytes_count(final_jmp))
        if len(validate_type) > 0:
            final_instructions += (validate_type + final_jmp)

            bytes_size = get_bytes_count(final_instructions)
            jmp_opcode, jmp_arg = OpcodeHelper.get_jump_and_data(Opcode.JMPIFNOT, bytes_size + 1)

            final_instructions = (is_type_opcodes +
                                  [(jmp_opcode, jmp_arg)] +
                                  final_instructions +
                                  return_false)
        else:
            is_type_opcodes.remove((Opcode.DUP, b''))
            final_instructions = is_type_opcodes

        return final_instructions

    def _is_instance_inner_opcodes(self, jmp_to_if_false: int = 0) -> List[Tuple[Opcode, bytes]]:
        if self.stack_item not in (StackItemType.Array, StackItemType.Struct, StackItemType.Map):
            return []

        variables_count_is_equal = [
            (Opcode.DUP, b''),
            (Opcode.SIZE, b''),
            OpcodeHelper.get_push_and_data(len(self._all_variables)),
            (Opcode.NUMEQUAL, b''),
        ]

        total_size = jmp_to_if_false
        variables_type_validations = []

        for var_index, (var_id, var) in reversed(list(enumerate(self._all_variables.items()))):

            if var.type.stack_item != StackItemType.Any:

                # validate primitive types only to avoid recursive code
                if self.stack_item == StackItemType.Map:
                    get = OpcodeHelper.get_pushdata_and_data(var_id)
                else:
                    get = OpcodeHelper.get_push_and_data(var_index)

                new_opcodes = [
                    (Opcode.DUP, b''),
                    get,
                    (Opcode.PICKITEM, b''),
                    (Opcode.ISTYPE, var.type.stack_item)
                ]
                if len(variables_type_validations) == 0:
                    new_opcodes.append((Opcode.NIP, b''))
                elif total_size > 0:
                    new_opcodes.append((OpcodeHelper.get_jump_and_data(Opcode.JMPIFNOT, total_size + 1)))

                variables_type_validations = new_opcodes + variables_type_validations
                total_size += get_bytes_count(new_opcodes)

        if total_size > 0:
            if len(variables_type_validations) > 0:
                final_opcode = (OpcodeHelper.get_jump_and_data(Opcode.JMPIFNOT, total_size + 1))
            else:
                final_opcode = (Opcode.NIP, b'')
            variables_count_is_equal.append(final_opcode)

        return variables_count_is_equal + variables_type_validations



from test_runner.blockchain.contract import TestRunnerContract as Contract
from test_runner.blockchain.log import TestRunnerLog as Log
from test_runner.blockchain.notification import TestRunnerNotification as Notification
from test_runner.blockchain.storage import TestRunnerStorage as Storage



from typing import Any, Dict, Optional, Sequence

from boa3.model.builtin.interop.interopmethod import InteropMethod
from boa3.model.builtin.interop.iterator.iteratortype import IteratorType
from boa3.model.builtin.method.builtinmethod import IBuiltinMethod
from boa3.model.expression import IExpression
from boa3.model.variable import Variable


class IteratorMethod(InteropMethod):

    def __init__(self, return_type: IteratorType):
        identifier = '-Iterator__init__'
        syscall = 'System.Iterator.Create'
        from boa3.model.type.type import Type
        args: Dict[str, Variable] = {'entry': Variable(Type.collection)}
        super().__init__(identifier, syscall, args, return_type=return_type)

    def validate_parameters(self, *params: IExpression) -> bool:
        return len(params) == 0

    @property
    def _args_on_stack(self) -> int:
        return len(self.args)

    @property
    def _body(self) -> Optional[str]:
        return

    def _entry_arg(self) -> Variable:
        return self.args['entry']

    def build(self, value: Any) -> IBuiltinMethod:
        if isinstance(value, Sequence) and len(value) == 1:
            argument = value[0]
        else:
            argument = value

        if self._entry_arg().type.is_type_of(argument):
            iterator = IteratorType.build(argument)
            if iterator is not self.return_type:
                return IteratorMethod(iterator)
        return super().build(value)



from __future__ import annotations

from boa3.builtin.interop.storage.storagemap import StorageMap
from boa3.builtin.type import ByteString


class StorageContext:
    """
    The storage context used to read and write data in smart contracts.
    """

    def __init__(self):
        pass

    def create_map(self, prefix: ByteString) -> StorageMap:
        """
        Creates a storage map with the given prefix.

        :param prefix: the identifier of the storage map
        :type prefix: str or bytes
        :return: a map with the key-values in the storage that match with the given prefix
        :rtype: StorageMap
        """
        pass

    def as_read_only(self) -> StorageContext:
        """
        Converts the specified storage context to a new readonly storage context.

        :return: current StorageContext as ReadOnly
        :rtype: StorageContext
        """
        pass



from typing import Any, List

import boa3_test.test_sc.import_test.FromImportTyping as UserModule
from boa3.builtin.compile_time import public


@public
def Main() -> List[Any]:
    return UserModule.EmptyList()



from __future__ import annotations

from typing import Dict

from boa3.model.symbol import ISymbol


class SymbolScope:
    def __init__(self, symbols: Dict[str, ISymbol] = None):
        self._symbols = symbols.copy() if symbols is not None else {}

    @property
    def symbols(self) -> Dict[str, ISymbol]:
        return self._symbols.copy()

    def copy(self) -> SymbolScope:
        return SymbolScope(self._symbols)

    def include_symbol(self, symbol_id: str, symbol: ISymbol, reassign_original: bool = True):
        """
        Includes a symbols into the scope

        :param symbol_id: symbol identifier
        :param symbol: symbol to be included
        """

        if symbol_id in self._symbols:
            if not reassign_original:
                if hasattr(symbol, 'copy'):
                    symbol = symbol.copy()
                else:
                    return

            if hasattr(symbol, 'set_is_reassigned'):
                symbol.set_is_reassigned()

        self._symbols[symbol_id] = symbol

    def remove_symbol(self, symbol_id: str):
        """
        Removes a symbols from the scope

        :param symbol_id: symbol identifier
        """
        if symbol_id in self._symbols:
            self._symbols.pop(symbol_id)

    def __getitem__(self, item: str) -> ISymbol:
        return self._symbols[item]

    def __contains__(self, item: str) -> bool:
        return item in self._symbols

    def __repr__(self) -> str:
        return self._symbols.__repr__()



import ast
from typing import Dict

from boa3.model.builtin.interop.contract.contracttype import ContractType
from boa3.model.builtin.interop.nativecontract import ContractManagementMethod
from boa3.model.variable import Variable


class CreateMethod(ContractManagementMethod):

    def __init__(self, contract_type: ContractType):
        from boa3.model.type.type import Type
        identifier = 'create_contract'
        syscall = 'deploy'
        args: Dict[str, Variable] = {
            'nef_file': Variable(Type.bytes),
            'manifest': Variable(Type.bytes),
            'data': Variable(Type.any)
        }
        data_default = ast.parse("{0}".format(Type.any.default_value)
                                 ).body[0].value

        super().__init__(identifier, syscall, args, defaults=[data_default], return_type=contract_type)



from __future__ import annotations

from typing import Dict, List, Optional

from boa3.model.identifiedsymbol import IdentifiedSymbol


class Package(IdentifiedSymbol):
    """
    A class used to represent a Python module.

    :ivar properties: a list that stores every property in the package. Empty by default.
    :ivar methods: a list that stores every method in the package. Empty by default.
    :ivar types: a list that stores every property in the package. Empty by default.
    """

    def __init__(self, identifier: str,
                 properties: List[IdentifiedSymbol] = None,
                 methods: List[IdentifiedSymbol] = None,
                 types: List[IdentifiedSymbol] = None,
                 packages: List[Package] = None,
                 other_symbols: dict = None,
                 import_origin=None
                 ):
        """
        :param packages: a list that stores the inner packages and modules of this package. Empty by default.
        :param other_symbols: a dictionary with other symbols that are evaluated during the compilation. Empty by
          default. Should be None for builtin packages.
        :type other_symbols: dict or None
        :param import_origin: the analyser that generated this package. Should be None for builtin packages.
        :type import_origin: boa3.model.imports.importsymbol.Import
        """

        from enum import Enum
        if isinstance(identifier, Enum):
            identifier = identifier.value

        super().__init__(identifier)

        self._all_symbols: List[IdentifiedSymbol] = []

        if methods is None:
            methods = []
        self._all_symbols.extend(methods)

        if properties is None:
            properties = []
        self._all_symbols.extend(properties)

        if types is None:
            types = []
        self._all_symbols.extend(types)

        if packages is None:
            packages = []
        self._packages = packages
        for package in packages:
            package._parent = self

        # for packages that are not builtin
        if isinstance(other_symbols, dict):
            from boa3.model.symbol import ISymbol
            self._additional_symbols = {key: value
                                        for key, value in other_symbols.items()
                                        if (value not in self._all_symbols
                                            and isinstance(key, str)
                                            and isinstance(value, ISymbol))}
            self.origin = import_origin
        else:
            self._additional_symbols = {}
            self.origin = None

        self._aliases: Dict[str, str] = {}
        self._parent: Optional[Package] = None

    @property
    def shadowing_name(self) -> str:
        return 'package'

    @property
    def symbols(self) -> Dict[str, IdentifiedSymbol]:
        """
        Gets all the symbols in the package.

        :return: a list that stores every symbol in the package
        """
        symbol_map = {(self._aliases[symbol.raw_identifier]
                       if symbol.raw_identifier in self._aliases
                       else symbol.raw_identifier): symbol
                      for symbol in self._all_symbols}
        symbol_map.update(self._additional_symbols)
        return symbol_map

    @property
    def inner_packages(self) -> Dict[str, Package]:
        return {symbol.raw_identifier: symbol for symbol in self._packages}

    @property
    def parent(self) -> Optional[Package]:
        """
        Get the parent package of this one. None if it's the root package.
        """
        return self._parent

    def include_symbol(self, symbol_id, symbol: IdentifiedSymbol):
        identifier = symbol.raw_identifier if isinstance(symbol, IdentifiedSymbol) else symbol_id
        check_list = self._packages if isinstance(symbol, Package) else self._all_symbols

        if all(package_symbol.raw_identifier != identifier for package_symbol in check_list):
            if isinstance(symbol, IdentifiedSymbol) and symbol.raw_identifier != symbol_id:
                self._aliases[symbol.raw_identifier] = symbol_id

            if isinstance(symbol, Package):
                self._packages.append(symbol)
                if symbol._parent is not None:
                    return
                symbol._parent = self
            else:
                self._all_symbols.append(symbol)

    def __repr__(self) -> str:
        return self.identifier



import ast
from abc import ABC
from typing import Dict, List, Optional, Tuple

from boa3.model.builtin.method.builtinmethod import IBuiltinMethod
from boa3.model.expression import IExpression
from boa3.model.type.itype import IType
from boa3.model.variable import Variable
from boa3.neo.vm.opcode.Opcode import Opcode


class InteropMethod(IBuiltinMethod, ABC):

    def __init__(self, identifier: str, sys_call: str, args: Dict[str, Variable] = None,
                 defaults: List[ast.AST] = None, return_type: IType = None):
        self._sys_call: str = sys_call
        super().__init__(identifier, args, defaults, return_type)

    def validate_parameters(self, *params: IExpression) -> bool:
        if any(not isinstance(param, IExpression) for param in params):
            return False

        args: List[IType] = [arg.type for arg in self.args.values()]
        if len(params) != len(args):
            return False
        return all(args[index].is_type_of(params[index]) for index in range(len(args)))

    @property
    def interop_method_hash(self) -> bytes:
        return self._method_hash(self._sys_call)

    def _method_hash(self, method_name: str) -> bytes:
        from boa3.constants import SIZE_OF_INT32
        from boa3.neo import cryptography
        from boa3.neo.vm.type.String import String

        return cryptography.sha256(String(method_name).to_bytes())[:SIZE_OF_INT32]

    @property
    def _opcode(self) -> List[Tuple[Opcode, bytes]]:
        return [(Opcode.SYSCALL, self.interop_method_hash)]

    @property
    def _args_on_stack(self) -> int:
        return len(self.args)

    @property
    def _body(self) -> Optional[str]:
        return None



import ast
from typing import Any, Dict, List, Optional, Tuple

from boa3.model.builtin.method.builtinmethod import IBuiltinMethod
from boa3.model.type.primitive.ibytestringtype import IByteStringType
from boa3.model.variable import Variable
from boa3.neo.vm.opcode import OpcodeHelper
from boa3.neo.vm.opcode.Opcode import Opcode


class StartsWithMethod(IBuiltinMethod):
    def __init__(self, self_type: IByteStringType = None):
        from boa3.model.type.type import Type

        if not isinstance(self_type, IByteStringType):
            from boa3.model.type.primitive.bytestringtype import ByteStringType
            self_type = ByteStringType.build()

        identifier = 'startswith'
        args: Dict[str, Variable] = {
            'self': Variable(self_type),
            'value': Variable(self_type),
            'start': Variable(Type.int),
            'end': Variable(Type.int),
        }

        start_default = ast.parse("{0}".format(0)
                                  ).body[0].value
        end_default = ast.parse("-1").body[0].value.operand
        end_default.n = -1

        super().__init__(identifier, args, defaults=[start_default, end_default], return_type=Type.bool)

    @property
    def _arg_self(self) -> Variable:
        return self.args['self']

    @property
    def _opcode(self) -> List[Tuple[Opcode, bytes]]:
        from boa3.compiler.codegenerator import get_bytes_count
        from boa3.neo.vm.type.StackItem import StackItemType

        jmp_place_holder = (Opcode.JMP, b'\x01')
        # string, end, start, substring

        verify_negative_index = [           # verify if index is negative value
            (Opcode.DUP, b''),
            (Opcode.PUSHM1, b''),
            jmp_place_holder                # if not index < 0, jump to verify_big_end
        ]

        fix_negative_end = [                # fix negative value to a positive equivalent
            (Opcode.PUSH3, b''),
            (Opcode.PICK, b''),
            (Opcode.SIZE, b''),
            (Opcode.ADD, b''),
            (Opcode.INC, b''),              # end = end + len(string) + 1
            (Opcode.DUP, b''),
            (Opcode.PUSHM1, b''),
            jmp_place_holder                # if not end < 0, jump to verify_and_fix_start
        ]

        fix_still_negative_index = [        # if end still is less than 0, then it should be 0
            (Opcode.DROP, b''),
            (Opcode.PUSH0, b''),            # end = 0
        ]

        jmp_fix_negative_index = OpcodeHelper.get_jump_and_data(Opcode.JMPGT, get_bytes_count(fix_negative_end +
                                                                                              fix_still_negative_index), True)
        verify_negative_index[-1] = jmp_fix_negative_index

        verify_big_end = [                  # verify if end is greater or equals to len(string)
            (Opcode.DUP, b''),
            (Opcode.PUSH4, b''),
            (Opcode.PICK, b''),
            (Opcode.SIZE, b''),
            jmp_place_holder                # if not end >= len(string), jump to verify_and_fix_start
        ]

        fix_big_end = [                     # fix end value to a positive equivalent
            (Opcode.DROP, b''),
            (Opcode.PUSH2, b''),
            (Opcode.PICK, b''),
            (Opcode.SIZE, b''),             # end = len(string)
        ]

        jmp_other_verifies = OpcodeHelper.get_jump_and_data(Opcode.JMPGT, get_bytes_count(fix_still_negative_index +
                                                                                          verify_big_end +
                                                                                          fix_big_end), True)
        fix_negative_end[-1] = jmp_other_verifies

        jmp_fix_big_index = OpcodeHelper.get_jump_and_data(Opcode.JMPLE, get_bytes_count(fix_big_end), True)
        verify_big_end[-1] = jmp_fix_big_index

        verify_and_fix_end = [              # verify and fix end index
            (Opcode.REVERSE3, b''),         # change positions on stack
        ]
        verify_and_fix_end.extend(verify_negative_index)
        verify_and_fix_end.extend(fix_negative_end)
        verify_and_fix_end.extend(fix_still_negative_index)
        verify_and_fix_end.extend(verify_big_end)
        verify_and_fix_end.extend(fix_big_end)

        verify_and_fix_start = [            # verify and fix end index
            (Opcode.SWAP, b''),             # change positions on stack
        ]

        # fix_negative_start is the same as fix_negative_end, but jump is different
        fix_negative_start = fix_negative_end.copy()
        verify_big_start = [                # verify if start is greater or equals to len(string)
            (Opcode.DUP, b''),
            (Opcode.PUSH4, b''),
            (Opcode.PICK, b''),
            (Opcode.SIZE, b''),
            jmp_place_holder                # if start >= len(string), return False
        ]

        jmp_other_verifies = OpcodeHelper.get_jump_and_data(Opcode.JMPGT, get_bytes_count(fix_still_negative_index +
                                                                                          verify_big_start), True)
        fix_negative_end[-1] = jmp_other_verifies

        verify_size = [                     # verify if len(string[start:end]) > len(substring)
            (Opcode.REVERSE3, b''),
            (Opcode.REVERSE4, b''),
            (Opcode.REVERSE3, b''),
            (Opcode.OVER, b''),
            (Opcode.SUB, b''),
            (Opcode.SUBSTR, b''),

            (Opcode.PUSH0, b''),
            (Opcode.PUSH2, b''),
            (Opcode.PICK, b''),
            (Opcode.SIZE, b''),
            (Opcode.DUP, b''),
            (Opcode.PUSH3, b''),
            (Opcode.PICK, b''),
            (Opcode.SIZE, b''),
            jmp_place_holder                # if len(string[start:end]) > len(substring), return False
        ]

        compare_starts = [                  # string.startswith(substring)
            (Opcode.SUBSTR, b''),
            (Opcode.CONVERT, StackItemType.ByteString),
            (Opcode.NUMEQUAL, b''),         # return string[start:end][0:len(substring)] == substring
            jmp_place_holder                # jumps other opcodes
        ]

        jmp_compare_starts = OpcodeHelper.get_jump_and_data(Opcode.JMPGT, get_bytes_count(compare_starts), True)
        verify_size[-1] = jmp_compare_starts

        jmp_to_false = OpcodeHelper.get_jump_and_data(Opcode.JMPGE, get_bytes_count(verify_size +
                                                                                    compare_starts), True)
        verify_big_start[-1] = jmp_to_false

        verify_and_fix_start.extend(verify_negative_index)
        verify_and_fix_start.extend(fix_negative_start)
        verify_and_fix_start.extend(fix_still_negative_index)
        verify_and_fix_start.extend(verify_big_start)

        return_false = [                    # remove all values from stack and put False
            (Opcode.DROP, b''),
            (Opcode.DROP, b''),
            (Opcode.DROP, b''),
            (Opcode.DROP, b''),
            (Opcode.PUSH0, b''),            # return False
        ]

        jmp_bigger_substr = OpcodeHelper.get_jump_and_data(Opcode.JMP, get_bytes_count(return_false), True)
        compare_starts[-1] = jmp_bigger_substr

        return (
            verify_and_fix_end +
            verify_and_fix_start +
            verify_size +
            compare_starts +
            return_false
        )

    def push_self_first(self) -> bool:
        return self.has_self_argument

    @property
    def _args_on_stack(self) -> int:
        return len(self.args)

    @property
    def _body(self) -> Optional[str]:
        return None

    def build(self, value: Any) -> IBuiltinMethod:
        if isinstance(value, IByteStringType):
            return StartsWithMethod(value)
        return super().build(value)



from boa3.exception import CompilerError
from boa3_test.tests.boa_test import BoaTest
from boa3_test.tests.test_classes.testengine import TestEngine


class TestIteratorInterop(BoaTest):
    default_folder: str = 'test_sc/interop_test/iterator'

    def test_iterator_create(self):
        path = self.get_contract_path('IteratorCreate.py')
        self.assertCompilerLogs(CompilerError.UnresolvedReference, path)

    def test_iterator_next(self):
        path = self.get_contract_path('IteratorNext.py')
        engine = TestEngine()

        prefix = 'test_iterator_next'
        result = self.run_smart_contract(engine, path, 'has_next', prefix)
        self.assertEqual(False, result)

        engine.storage_put(prefix + 'example1', 1, contract_path=path)
        result = self.run_smart_contract(engine, path, 'has_next', prefix)
        self.assertEqual(True, result)

    def test_iterator_value(self):
        path = self.get_contract_path('IteratorValue.py')
        engine = TestEngine()

        prefix = 'test_iterator_value'
        result = self.run_smart_contract(engine, path, 'test_iterator', prefix)
        self.assertIsNone(result)

        key = prefix + 'example1'
        engine.storage_put(key, 1, contract_path=path)
        result = self.run_smart_contract(engine, path, 'test_iterator', prefix)
        self.assertEqual([key, '\x01'], result)

    def test_iterator_value_dict_mismatched_type(self):
        path = self.get_contract_path('IteratorValueMismatchedType.py')
        self.assertCompilerLogs(CompilerError.MismatchedTypes, path)

    def test_import_iterator(self):
        path = self.get_contract_path('ImportIterator.py')
        engine = TestEngine()

        result = self.run_smart_contract(engine, path, 'return_iterator')
        self.assertEqual([], result)

    def test_import_interop_iterator(self):
        path = self.get_contract_path('ImportInteropIterator.py')
        engine = TestEngine()

        result = self.run_smart_contract(engine, path, 'return_iterator')
        self.assertEqual([], result)

    def test_iterator_implicit_typing(self):
        path = self.get_contract_path('IteratorImplicitTyping.py')
        engine = TestEngine()

        prefix = 'test_iterator_'
        result = self.run_smart_contract(engine, path, 'search_storage', prefix)
        self.assertEqual({}, result)

        result = self.run_smart_contract(engine, path, 'store', f'{prefix}1', 1)
        self.assertIsVoid(result)

        result = self.run_smart_contract(engine, path, 'store', f'{prefix}2', 2)
        self.assertIsVoid(result)

        result = self.run_smart_contract(engine, path, 'search_storage', prefix)
        self.assertEqual({f'{prefix}1': 1, f'{prefix}2': 2}, result)

    def test_iterator_value_access(self):
        path = self.get_contract_path('IteratorValueAccess.py')
        engine = TestEngine()

        prefix = 'test_iterator_'
        result = self.run_smart_contract(engine, path, 'search_storage', prefix)
        self.assertEqual({}, result)

        result = self.run_smart_contract(engine, path, 'store', f'{prefix}1', 1)
        self.assertIsVoid(result)

        result = self.run_smart_contract(engine, path, 'store', f'{prefix}2', 2)
        self.assertIsVoid(result)

        result = self.run_smart_contract(engine, path, 'search_storage', prefix)
        self.assertEqual({f'{prefix}1': 1, f'{prefix}2': 2}, result)



from typing import Dict, List

from boa3.model.callable import Callable
from boa3.model.identifiedsymbol import IdentifiedSymbol
from boa3.model.symbol import ISymbol
from boa3.model.type.annotation.metatype import metaType
from boa3.model.type.type import Type
from boa3.model.type.typingmethod.casttypemethod import CastTypeMethod


class TypeUtils:
    @classmethod
    def all_functions(cls) -> Dict[str, Callable]:
        from boa3.model.builtin.builtincallable import IBuiltinCallable
        return {tpe._identifier: tpe for tpe in vars(cls).values() if isinstance(tpe, IBuiltinCallable)}

    @classmethod
    def get_types_from_typing_lib(cls) -> Dict[str, ISymbol]:
        import typing
        from types import FunctionType

        type_symbols: Dict[str, ISymbol] = {}
        all_types: List[str] = typing.__all__

        for t_id in all_types:
            attr = getattr(typing, t_id)
            if not isinstance(attr, FunctionType):
                type_id: str = t_id if t_id in Type.all_types() else t_id.lower()
                if type_id in Type.all_types():
                    type_symbols[t_id] = Type.all_types()[type_id]
            else:
                function_id: str = t_id if t_id in cls.all_functions() else t_id.lower()
                if function_id in cls.all_functions():
                    type_symbols[t_id] = cls.all_functions()[function_id]

        return type_symbols

    # type for internal validation
    type = metaType

    # Annotation function utils
    cast = CastTypeMethod()

    _internal_validation_symbols: List[IdentifiedSymbol] = [type
                                                            ]

    symbols_for_internal_validation = {symbol.identifier: symbol
                                       for symbol in _internal_validation_symbols}



from typing import Dict, List, Tuple

from boa3.model.builtin.interop.nativecontract import CryptoLibMethod
from boa3.model.variable import Variable
from boa3.neo.vm.opcode.Opcode import Opcode


class Hash256Method(CryptoLibMethod):
    def __init__(self):
        from boa3.model.type.type import Type
        identifier = 'hash256'
        native_identifier = ''  # hash256 is not neo native
        args: Dict[str, Variable] = {'key': Variable(Type.any)}
        super().__init__(identifier, native_identifier, args, return_type=Type.bytes)

    @property
    def pack_arguments(self) -> bool:
        if self._pack_arguments is None:
            from boa3.model.builtin.interop.interop import Interop
            self._pack_arguments = Interop.Sha256.pack_arguments  # this is the first method called
        return self._pack_arguments

    @property
    def _opcode(self) -> List[Tuple[Opcode, bytes]]:
        # calls the generic implementation to ensure the correct output when calling consecutive compilations
        default_opcodes = super()._opcode
        from boa3.model.builtin.interop.interop import Interop
        return (Interop.Sha256.opcode
                + Interop.Sha256.opcode)



import ast
from typing import Dict, Optional

from boa3.model.builtin.classmethod.popmethod import PopMethod
from boa3.model.type.itype import IType
from boa3.model.variable import Variable


class PopSequenceMethod(PopMethod):

    def __init__(self, arg_value: Optional[IType] = None):
        from boa3.model.type.type import Type

        if not Type.mutableSequence.is_type_of(arg_value):
            arg_value = Type.mutableSequence

        args: Dict[str, Variable] = {
            'self': Variable(arg_value),
            'index': Variable(arg_value.valid_key)
        }

        index_default = ast.parse("-1").body[0].value.operand
        index_default.n = -1

        super().__init__(args, defaults=[index_default], return_type=arg_value.value_type)



from enum import IntFlag


class OracleResponseCode(IntFlag):
    """
    Represents the response code for the oracle request.
    """

    SUCCESS = 0x00
    """
    Indicates that the request has been successfully completed.

    :meta hide-value:
    """

    PROTOCOL_NOT_SUPPORTED = 0x10
    """
    Indicates that the protocol of the request is not supported.

    :meta hide-value:
    """

    CONSENSUS_UNREACHABLE = 0x12
    """
    Indicates that the oracle nodes cannot reach a consensus on the result of the request.

    :meta hide-value:
    """

    NOT_FOUND = 0x14
    """
    Indicates that the requested Uri does not exist.

    :meta hide-value:
    """

    TIME_OUT = 0x16
    """
    Indicates that the request was not completed within the specified time.

    :meta hide-value:
    """

    FORBIDDEN = 0x18
    """
    Indicates that there is no permission to request the resource.

    :meta hide-value:
    """

    RESPONSE_TOO_LARGE = 0x1A
    """
    Indicates that the data for the response is too large.

    :meta hide-value:
    """

    INSUFFICIENT_FUNDS = 0x1C
    """
    Indicates that the request failed due to insufficient balance.

    :meta hide-value:
    """

    CONTENT_TYPE_NOT_SUPPORTED = 0x1F
    """
    Indicates that the content-type of the request is not supported.

    :meta hide-value:
    """

    ERROR = 0xFF
    """
    Indicates that the request failed due to other errors.

    :meta hide-value:
    """



from typing import Optional

from boa3 import constants
from boa3.model.identifiedsymbol import IdentifiedSymbol
from boa3.model.method import Method
from boa3.model.type.classes.userclass import UserClass
from boa3.model.variable import Variable


class ClassInitMethod(IdentifiedSymbol, Method):
    def __init__(self, user_class: UserClass):
        self_var = Variable(user_class)
        args = {
            'self': self_var
        }
        base_init = None
        if len(user_class.bases) == 1:
            # TODO: change when class inheritance with multiple bases is implemented
            # use update to keep the original order
            base_init = user_class.bases[0].constructor_method()
            args.update(base_init.args)
            # but change the self type
            args['self'] = self_var

        Method.__init__(self, args=args, return_type=user_class)
        IdentifiedSymbol.__init__(self, identifier=constants.INIT_METHOD_ID)

        self.defined_by_entry = False
        self.is_init = True
        self.origin_class = user_class
        # __init__ method behave like class methods
        from boa3.model.builtin.builtin import Builtin
        self.decorators.append(Builtin.ClassMethodDecorator)

        if base_init is not None:
            origin = self.origin if self.origin else self
            base_init.add_call_origin(origin)

    @property
    def _args_on_stack(self) -> int:
        return len(self.args)

    @property
    def _body(self) -> Optional[str]:
        return None



from enum import IntFlag


class Role(IntFlag):
    """
    Represents the roles in the NEO system. They are the permission types of the native contract `RoleManagement`.
    """

    STATE_VALIDATOR = 4
    """
    The validators of state. Used to generate and sign the state root.

    :meta hide-value:
    """

    ORACLE = 8
    """
    The nodes used to process Oracle requests.

    :meta hide-value:
    """

    NEO_FS_ALPHABET_NODE = 16
    """
    NeoFS Alphabet nodes.

    :meta hide-value:
    """



from __future__ import annotations

import ast
from typing import Optional

from boa3.model.expression import IExpression
from boa3.model.identifiedsymbol import IdentifiedSymbol
from boa3.model.symbol import ISymbol
from boa3.model.type.itype import IType


class GeneratorData:

    def __init__(self, origin_node: Optional[ast.AST],
                 symbol_id: Optional[str] = None,
                 symbol: Optional[ISymbol] = None,
                 result_type: Optional[IType] = None,
                 index: Optional[int] = None,
                 origin_object_type: Optional[IType] = None,
                 already_generated: bool = False):

        if symbol_id is None and isinstance(symbol, IdentifiedSymbol):
            symbol_id = symbol.identifier
        if result_type is None and isinstance(symbol, IExpression):
            result_type = symbol.type

        self.node: Optional[ast.AST] = origin_node
        self.symbol: Optional[ISymbol] = symbol
        self.symbol_id: Optional[str] = symbol_id
        self.type: Optional[IType] = result_type
        self.index: Optional[int] = index
        self.origin_object_type: Optional[ISymbol] = origin_object_type
        self.already_generated: bool = already_generated

    def copy(self, new_origin: Optional[ast.AST] = None) -> GeneratorData:
        return GeneratorData(new_origin if isinstance(new_origin, ast.AST) else self.node,
                             self.symbol_id,
                             self.symbol,
                             self.type,
                             self.index,
                             self.already_generated)



from abc import ABC
from typing import Iterable, Optional, Union

from boa3 import constants
from boa3.model.builtin.internal.internalmethod import IInternalMethod
from boa3.model.event import Event
from boa3.model.method import Method


class CompilerError(ABC, BaseException):
    """
    An interface for compilation errors
    """

    def __init__(self, line: int, col: int):
        self.line: int = line
        self.col: int = col
        self.filepath: Optional[str] = None

    @property
    def message(self) -> str:
        message = '' if self._error_message is None else ' - ' + self._error_message
        if isinstance(self.filepath, str):
            message += f'\t <{self.filepath}>'
        return '{0}:{1}{2}'.format(self.line, self.col, message)

    @property
    def _error_message(self) -> Optional[str]:
        return None

    def __str__(self) -> str:
        return self.message

    def __eq__(self, other) -> bool:
        if not isinstance(other, type(self)):
            return False
        return self.message == other.message


class CircularImport(CompilerError):
    """
    An error raised when circular imports are detected
    """

    def __init__(self, line: int, col: int, target_import: str, target_origin: str):
        import os
        self.target_import = target_import
        self.target_origin = target_origin.replace(os.sep, constants.PATH_SEPARATOR)
        super().__init__(line, col)

    @property
    def _error_message(self) -> Optional[str]:
        return "Circular import with '%s' ('%s')" % (self.target_import, self.target_origin)


class DuplicatedIdentifier(CompilerError):
    """
    An error raised when more than one symbol uses the same identifier in the same scope and cannot be overwritten.
    """

    def __init__(self, line: int, col: int, duplicated_id: str = None):
        self._duplicated_id = duplicated_id
        super().__init__(line, col)

    @property
    def _error_message(self) -> Optional[str]:
        return f"Duplicate identifier: '{self._duplicated_id}'"


class DuplicatedManifestIdentifier(CompilerError):
    """
    An error raised when more than one symbol uses the same identifier in the manifest.
    """

    def __init__(self, line: int, col: int, duplicated_id: str = None, duplicated_arg_count: int = None):
        self._duplicated_id = duplicated_id
        self._arg_count = duplicated_arg_count
        super().__init__(line, col)

    @property
    def _error_message(self) -> Optional[str]:
        return f"Duplicate manifest identifier: '{self._duplicated_id}' with {self._arg_count} argument(s)"


class IncorrectNumberOfOperands(CompilerError):
    """
    An error raised when an operation is used with the wrong number of operands
    """

    def __init__(self, line: int, col: int, expected_count: int, actual_count: int):
        self.expected = expected_count
        self.actual = actual_count
        super().__init__(line, col)

    @property
    def _error_message(self) -> Optional[str]:
        return "Incorrect number of operands: expected '%s', got '%s' instead" % (self.expected, self.actual)


class InvalidUsage(CompilerError):
    """
    An error raised when a built-in function or decorator is incorrectly defined
    """

    def __init__(self, line: int, col: int, custom_error_message: str = None):
        self.custom_error_message = custom_error_message
        super().__init__(line, col)

    @property
    def _error_message(self) -> Optional[str]:
        message = "Invalid usage"
        if self.custom_error_message is not None:
            message += f": {self.custom_error_message}"
        return message


class InvalidType(CompilerError):
    """
    An error raised when a type that is not supported by Neo VM is used
    """

    def __init__(self, line: int, col: int, symbol_id: str = None):
        self.symbol_id = symbol_id
        super().__init__(line, col)

    @property
    def _error_message(self) -> Optional[str]:
        message = "Invalid type"
        if self.symbol_id is not None:
            message += ": '%s'" % self.symbol_id
        return message


class InternalError(CompilerError):
    """
    An error raised when an unexpected exception is raised during the compilation
    """

    def __init__(self, line: int, col: int, raised_exception: BaseException = None):
        self.raised_exception: BaseException = raised_exception
        super().__init__(line, col)

    @property
    def _error_message(self) -> Optional[str]:
        message = "Internal compiler error"
        if self.raised_exception is not None:
            message += ". {0}: {1}".format(type(self.raised_exception).__name__, str(self.raised_exception))
        return message


class InternalIncorrectSignature(CompilerError):
    """
    An error raised when an internal method is defined by the user, but with an incorrect signature
    """

    def __init__(self, line: int, col: int, expected_method: IInternalMethod):
        self.expected_method: IInternalMethod = expected_method
        super().__init__(line, col)

    @property
    def _error_message(self) -> Optional[str]:
        return "The implementation of '{0}' is different " \
               "from the expected '{1}'.".format(self.expected_method.raw_identifier,
                                                 self.expected_method)


class MetadataImplementationMissing(CompilerError):
    """
    An error raised when the metadata required functions aren't implemented
    """

    def __init__(self, line: int, col: int, symbol_id: str, metadata_attr_id: str):
        self.symbol_id: str = symbol_id
        self.metadata_attr_id: str = metadata_attr_id
        super().__init__(line, col)

    @property
    def _error_message(self) -> Optional[str]:
        return "'{0}' requires '{1}' implementation".format(self.metadata_attr_id, self.symbol_id)


class MetadataIncorrectImplementation(CompilerError):
    """
    An error raised when a metadata required function is incorrectly implemented
    """
    from boa3.model.symbol import ISymbol

    def __init__(self, line: int, col: int, symbol_id: str, expected_symbol: ISymbol, actual_symbol: ISymbol):
        from boa3.model.symbol import ISymbol

        self.symbol_id: str = symbol_id
        self.expected_symbol: ISymbol = expected_symbol
        self.actual_symbol: ISymbol = actual_symbol
        super().__init__(line, col)

    @property
    def _error_message(self) -> Optional[str]:
        return ("'{0}' is not correctly implemented. Expecting '{1}', got '{2}' instead"
                .format(self.symbol_id, self.expected_symbol, self.actual_symbol))


class MetadataInformationMissing(CompilerError):
    """
    An error raised when the metadata info doesn't match with the functions requirements
    """

    def __init__(self, line: int, col: int, symbol_id: str, metadata_attr_id: str):
        self.symbol_id: str = symbol_id
        self.metadata_attr_id: str = metadata_attr_id
        super().__init__(line, col)

    @property
    def _error_message(self) -> Optional[str]:
        return "'{0}' requires '{1}' attribute, which is missing in the metadata".format(self.symbol_id,
                                                                                         self.metadata_attr_id)


class MismatchedTypes(CompilerError):
    """
    An error raised when the evaluated and expected types are not the same
    """

    def __init__(self, line: int, col: int, expected_type_id: Union[str, Iterable[str]],
                 actual_type_id: Union[str, Iterable[str]]):
        if isinstance(expected_type_id, str):
            expected_type_id = [expected_type_id]
        if isinstance(actual_type_id, str):
            actual_type_id = [actual_type_id]

        self.expected_types = expected_type_id
        self.actual_types = actual_type_id
        super().__init__(line, col)

    @property
    def _error_message(self) -> Optional[str]:
        expected_types = join_args(self.expected_types)
        actual_types = join_args(self.actual_types)
        return "Expected type '%s', got '%s' instead" % (expected_types, actual_types)


class MissingInitCall(CompilerError):
    """
    An error raised when a custom class is created with inheritance and it's missing the base __init__ call
    """

    def __init__(self, line: int, col: int):
        super().__init__(line, col)

    @property
    def _error_message(self) -> Optional[str]:
        return "Call to __init__ of super class is missed"


class MissingReturnStatement(CompilerError):
    """
    An error raised when a function with a return value is missing a return statement
    """

    def __init__(self, line: int, col: int, symbol_id: str):
        self.symbol_id = symbol_id
        super().__init__(line, col)

    @property
    def _error_message(self) -> Optional[str]:
        return "'%s': Missing return statement" % self.symbol_id


class MissingStandardDefinition(CompilerError):
    """
    An error raised when a contract standard is defined in the metadata and are required symbols missing
    """

    def __init__(self, standard_id: str, symbol_id: str, symbol: Union[Method, Event]):
        self.standard = standard_id
        self.symbol_id = symbol_id
        self.symbol = symbol
        super().__init__(0, 0)

    @property
    def _error_message(self) -> Optional[str]:
        safe_symbol_prefix = 'safe ' if self.symbol.is_safe else ''
        return f"'{self.standard}': Missing '{self.symbol_id}' {self.symbol.shadowing_name} definition " \
               f"'{safe_symbol_prefix}{self.symbol}'"

    @property
    def message(self) -> str:
        return self._error_message


class NotSupportedOperation(CompilerError):
    """
    An error raised when an operation that is not supported by Neo VM is used
    """

    def __init__(self, line: int, col: int, symbol_id: str):
        self.symbol_id = symbol_id
        super().__init__(line, col)

    @property
    def _error_message(self) -> Optional[str]:
        return "The following operation is not supported: '%s'" % self.symbol_id


class UnexpectedArgument(CompilerError):
    """
    An error thrown when more arguments are used in a function than the number of arguments in the function's signature
    """

    def __init__(self, line: int, col: int):
        super().__init__(line, col)

    @property
    def _error_message(self) -> Optional[str]:
        return "Unexpected argument"


class UnfilledArgument(CompilerError):
    """
    An error thrown when less arguments are used in a function than the number of arguments in the function's signature
    """

    def __init__(self, line: int, col: int, param: str):
        self.param = param
        super().__init__(line, col)

    @property
    def _error_message(self) -> Optional[str]:
        return "Parameter '%s' unfilled" % self.param


class UnresolvedReference(CompilerError):
    """
    An error raised when an undefined symbol is used
    """

    def __init__(self, line: int, col: int, symbol_id: str):
        self.symbol_id = symbol_id
        super().__init__(line, col)

    @property
    def _error_message(self) -> Optional[str]:
        return "Unresolved reference '%s'" % self.symbol_id

    def __eq__(self, other) -> bool:
        if not isinstance(other, type(self)):
            return False
        return self._error_message == other._error_message


class UnresolvedOperation(CompilerError):
    """
    An error raised when an undefined symbol is used
    """

    def __init__(self, line: int, col: int, type_id: str, operation_id: str):
        self.type_id = type_id
        self.operation_id = operation_id
        super().__init__(line, col)

    @property
    def _error_message(self) -> Optional[str]:
        return "Unresolved reference: '{0}' does not have a definition of '{1}' operator".format(self.type_id,
                                                                                                 self.operation_id)


class TooManyReturns(CompilerError):
    """
    An error raised when a function returns a tuple
    """

    def __init__(self, line: int, col: int):
        super().__init__(line, col)

    @property
    def _error_message(self) -> Optional[str]:
        return "Too many returns"


class TypeHintMissing(CompilerError):
    """
    An error raised when type hint cannot be found
    """

    def __init__(self, line: int, col: int, symbol_id: str = None):
        self.symbol_id = symbol_id
        super().__init__(line, col)

    @property
    def _error_message(self) -> Optional[str]:
        if self.symbol_id is not None:
            return "Type hint is missing for the symbol '%s'" % self.symbol_id


class SelfArgumentError(CompilerError):
    """
    An error raised when the self argument is wrong
    """

    def __init__(self, line: int, col: int):
        super().__init__(line, col)

    @property
    def _error_message(self) -> Optional[str]:
        return "The self argument was not found or the annotation is incorrect"


def join_args(iterable: Iterable[str]) -> str:
    return str.join("', '", iterable)



from boa3.exception import CompilerError
from boa3.neo.vm.type.String import String
from boa3_test.tests.boa_test import BoaTest
from boa3_test.tests.test_classes.testengine import TestEngine


class TestReversed(BoaTest):
    default_folder: str = 'test_sc/reversed_test'

    def test_reversed_list_bool(self):
        path = self.get_contract_path('ReversedListBool.py')
        engine = TestEngine()

        list_bool = [True, True, False]
        result = self.run_smart_contract(engine, path, 'main')
        reversed_list = [element for element in reversed(list_bool)]
        self.assertEqual(reversed_list, result)

    def test_reversed_list_bytes(self):
        path = self.get_contract_path('ReversedListBytes.py')
        engine = TestEngine()

        list_bytes = [b'1', b'2', b'3']
        reversed_list = [element for element in reversed(list_bytes)]
        result = self.run_smart_contract(engine, path, 'main')
        if isinstance(result, list):
            for k in range(len(result)):
                if isinstance(result[k], str):
                    result[k] = String(result[k]).to_bytes()
        self.assertEqual(reversed_list, result)

    def test_reversed_list_int(self):
        path = self.get_contract_path('ReversedListInt.py')
        engine = TestEngine()

        list_int = [1, 2, 3]
        result = self.run_smart_contract(engine, path, 'main')
        reversed_list = [element for element in reversed(list_int)]
        self.assertEqual(reversed_list, result)

    def test_reversed_list_str(self):
        path = self.get_contract_path('ReversedListStr.py')
        engine = TestEngine()

        list_str = ['neo3-boa', 'unit', 'test']
        result = self.run_smart_contract(engine, path, 'main')
        reversed_list = [element for element in reversed(list_str)]
        self.assertEqual(reversed_list, result)

    def test_reversed_list(self):
        path = self.get_contract_path('ReversedList.py')
        engine = TestEngine()

        list_any = [1, 'string', False]
        result = self.run_smart_contract(engine, path, 'main', list_any)
        reversed_list = [element for element in reversed(list_any)]
        self.assertEqual(reversed_list, result)

    def test_reversed_string(self):
        path = self.get_contract_path('ReversedString.py')
        engine = TestEngine()

        string = 'unit_test'
        result = self.run_smart_contract(engine, path, 'main', string)
        reversed_list = [element for element in reversed(string)]
        self.assertEqual(reversed_list, result)

    def test_reversed_bytes(self):
        path = self.get_contract_path('ReversedBytes.py')
        engine = TestEngine()

        bytes_value = b'unit_test'
        reversed_list = [element for element in reversed(bytes_value)]
        result = self.run_smart_contract(engine, path, 'main', bytes_value)
        self.assertEqual(reversed_list, result)

    def test_reversed_range(self):
        path = self.get_contract_path('ReversedRange.py')
        engine = TestEngine()

        result = self.run_smart_contract(engine, path, 'main')
        reversed_list = [element for element in reversed(range(3))]
        self.assertEqual(reversed_list, result)

    def test_reversed_tuple(self):
        path = self.get_contract_path('ReversedTuple.py')
        engine = TestEngine()

        tuple_value = (1, 2, 3)
        result = self.run_smart_contract(engine, path, 'main', tuple_value)
        reversed_list = [element for element in reversed(tuple_value)]
        self.assertEqual(reversed_list, result)

    def test_mismatched_type(self):
        path = self.get_contract_path('ReversedParameterMismatchedType')
        self.assertCompilerLogs(CompilerError.MismatchedTypes, path)



from typing import Dict

from boa3.model.method import Method
from boa3.model.type.itype import IType
from boa3.model.type.type import Type
from boa3.model.variable import Variable


class StandardMethod(Method):
    def __init__(self, display_name: str,
                 args: Dict[str, IType] = None,
                 return_type: IType = Type.none,
                 safe: bool = False):
        if not isinstance(args, dict):
            args = {}
        method_args = {key: Variable(value) for key, value in args.items()}
        super().__init__(args=method_args, return_type=return_type, is_public=True,
                         external_name=display_name, is_safe=safe)



from typing import List

from boa3.model.callable import Callable
from boa3.model.type.classes.classtype import ClassType
from boa3.model.type.classes.contractinterfacehash import ContractHashProperty
from boa3.model.type.classes.userclass import UserClass
from boa3.neo3.core.types import UInt160


class ContractInterfaceClass(UserClass):
    def __init__(self, identifier: str, contract_hash: UInt160, decorators: List[Callable] = None,
                 bases: List[ClassType] = None):
        from boa3.model.builtin.builtin import Builtin, ContractDecorator

        if not isinstance(decorators, list):
            decorators = [Builtin.ContractInterface]
        elif not any(isinstance(decorator, ContractDecorator) for decorator in decorators):
            decorators.append(Builtin.ContractInterface)

        super().__init__(identifier, decorators, bases)
        self.contract_hash = contract_hash

        contract_hash_property = ContractHashProperty(f'-{identifier}_hash', contract_hash.to_array())
        self.include_property('hash', contract_hash_property)

    @property
    def is_interface(self) -> bool:
        return True



from typing import Any, List

from boa3.builtin.compile_time import public
from boa3_test.test_sc.import_test.FromImportTyping import EmptyList as NewList


@public
def Main() -> list:
    a: List[Any] = NewList()
    return a



from __future__ import annotations

from typing import Any, Dict

from boa3.model.builtin.interop.nativecontract import StdLibContract
from boa3.model.builtin.native.inativecontractclass import INativeContractClass
from boa3.model.method import Method


class StdLibClass(INativeContractClass):
    """
    A class used to represent StdLib native contract
    """

    def __init__(self):
        super().__init__('StdLib', StdLibContract)

    @property
    def class_methods(self) -> Dict[str, Method]:
        # avoid recursive import
        from boa3.model.builtin.interop.interop import Interop

        if len(self._class_methods) == 0:
            self._class_methods = {
                'serialize': Interop.Serialize,
                'deserialize': Interop.Deserialize,
                'json_serialize': Interop.JsonSerialize,
                'json_deserialize': Interop.JsonDeserialize,
                'base64_decode': Interop.Base64Decode,
                'base64_encode': Interop.Base64Encode,
                'base58_decode': Interop.Base58Decode,
                'base58_encode': Interop.Base58Encode,
                'base58_check_decode': Interop.Base58CheckDecode,
                'base58_check_encode': Interop.Base58CheckEncode,
                'itoa': Interop.Itoa,
                'atoi': Interop.Atoi,
                'memory_compare': Interop.MemoryCompare,
                'memory_search': Interop.MemorySearch,
            }
        return super().class_methods

    @classmethod
    def build(cls, value: Any = None) -> StdLibClass:
        if value is None or cls._is_type_of(value):
            return _StdLib

    @classmethod
    def _is_type_of(cls, value: Any):
        return isinstance(value, StdLibClass)


_StdLib = StdLibClass()



import ast
from typing import Dict, List, Optional, Tuple

from boa3.model.builtin.classmethod.indexmethod import IndexMethod
from boa3.model.expression import IExpression
from boa3.model.type.primitive.bytestype import BytesType
from boa3.model.type.primitive.strtype import StrType
from boa3.model.variable import Variable
from boa3.neo.vm.opcode import OpcodeHelper
from boa3.neo.vm.opcode.Opcode import Opcode
from boa3.neo.vm.type.Integer import Integer
from boa3.neo.vm.type.String import String


class IndexBytesStringMethod(IndexMethod):

    def __init__(self, self_type: Optional[StrType] = None):
        from boa3.model.type.type import Type
        if not isinstance(self_type, (StrType, BytesType)):
            self_type = Type.str

        args: Dict[str, Variable] = {
            'self': Variable(self_type),
            'x': Variable(self_type),
            'start': Variable(Type.int),
            'end': Variable(Type.int),
        }

        start_default = ast.parse("{0}".format(0)
                                  ).body[0].value
        end_default = ast.parse("-1").body[0].value.operand
        end_default.n = -1

        super().__init__(args, defaults=[start_default, end_default])

    def validate_parameters(self, *params: IExpression) -> bool:
        if 2 <= len(params) <= 4:
            return False
        if not all(isinstance(param, IExpression) for param in params):
            return False

        from boa3.model.type.itype import IType
        self_type: IType = params[0].type

        if not isinstance(self_type, (StrType, BytesType)):
            return False

        if not self_type.is_type_of(params[1]):
            return False

        return True

    @property
    def error_message(self) -> str:
        return 'substring not found' if isinstance(self._arg_self.type, StrType) else 'subsequence of bytes not found'

    @property
    def _opcode(self) -> List[Tuple[Opcode, bytes]]:
        from boa3.compiler.codegenerator import get_bytes_count
        from boa3.neo.vm.type.StackItem import StackItemType

        jmp_place_holder = (Opcode.JMP, b'\x01')
        message = String(self.error_message).to_bytes()

        # receives: end, start, substr, str
        verify_negative_index = [           # verifies if index in a negative value
            (Opcode.DUP, b''),
            (Opcode.PUSHM1, b''),
            jmp_place_holder                # if index >= 0, jump to verify_big_end or verify_big_start
        ]

        fix_negative_end = [                # gets the correspondent positive value of end
            (Opcode.PUSH3, b''),
            (Opcode.PICK, b''),
            (Opcode.SIZE, b''),
            (Opcode.ADD, b''),
            (Opcode.INC, b''),              # end = end + len(str) + 1
            (Opcode.DUP, b''),
            (Opcode.PUSHM1, b''),
            jmp_place_holder                # if end is not negative anymore, start verifying start
        ]

        fix_still_negative_index = [        # if index is still negative, consider it 0 then
            (Opcode.DROP, b''),
            (Opcode.PUSH0, b''),            # end = 0
        ]

        jmp_fix_negative_index = OpcodeHelper.get_jump_and_data(Opcode.JMPGT, get_bytes_count(fix_negative_end +
                                                                                              fix_still_negative_index), True)
        verify_negative_index[-1] = jmp_fix_negative_index

        verify_big_end = [                  # verify if end is bigger then len(str)
            (Opcode.DUP, b''),
            (Opcode.PUSH4, b''),
            (Opcode.PICK, b''),
            (Opcode.SIZE, b''),
            jmp_place_holder                # if end <= len(str), start verifying start
        ]

        fix_big_end = [                     # consider end as len(str)
            (Opcode.DROP, b''),
            (Opcode.PUSH2, b''),
            (Opcode.PICK, b''),
            (Opcode.SIZE, b''),             # end = len(str)
        ]

        jmp_other_verifies = OpcodeHelper.get_jump_and_data(Opcode.JMPGT, get_bytes_count(fix_still_negative_index +
                                                                                          verify_big_end +
                                                                                          fix_big_end), True)
        fix_negative_end[-1] = jmp_other_verifies

        jmp_fix_big_index = OpcodeHelper.get_jump_and_data(Opcode.JMPLE, get_bytes_count(fix_big_end), True)
        verify_big_end[-1] = jmp_fix_big_index

        verify_and_fix_end = [              # collection of Opcodes regarding verifying and fixing end index
            (Opcode.REVERSE4, b''),
        ]
        verify_and_fix_end.extend(verify_negative_index)
        verify_and_fix_end.extend(fix_negative_end)
        verify_and_fix_end.extend(fix_still_negative_index)
        verify_and_fix_end.extend(verify_big_end)
        verify_and_fix_end.extend(fix_big_end)

        verify_and_fix_start = [            # collection of Opcodes regarding verifying and fixing start index
            (Opcode.SWAP, b'')
        ]
        verify_and_fix_start.extend(verify_negative_index)
        verify_and_fix_start.extend(fix_negative_end)
        verify_and_fix_start.extend(fix_still_negative_index)
        verify_and_fix_start.extend(verify_big_end)
        verify_and_fix_start.extend(fix_big_end)

        change_stack_order = [              # change order of items on the stack
            (Opcode.PUSH2, b''),
            (Opcode.PICK, b''),
            (Opcode.SIZE, b''),             # size = len(substr)
            (Opcode.ROT, b''),
            (Opcode.ROT, b''),
        ]

        verify_while = [                    # verify already compared all that was need
            (Opcode.OVER, b''),
            (Opcode.OVER, b''),
            (Opcode.SUB, b''),
            (Opcode.PUSH3, b''),
            (Opcode.PICK, b''),
            jmp_place_holder                # if end - index >= size, jump to not_inside_sequence
        ]

        compare_item = [                    # compare str[index:index+size] with substr
            (Opcode.PUSH4, b''),
            (Opcode.PICK, b''),
            (Opcode.OVER, b''),
            (Opcode.PUSH4, b''),
            (Opcode.PICK, b''),
            (Opcode.SUBSTR, b''),
            (Opcode.CONVERT, StackItemType.ByteString),
            (Opcode.PUSH4, b''),
            (Opcode.PICK, b''),
            (Opcode.NUMEQUAL, b''),
            jmp_place_holder                # if str[index:index+size] == substr, return index
        ]

        not_found = [                       # increments index and goes back to verify again
            (Opcode.INC, b''),              # index++
            # jump to verify_while
        ]

        jmp_back_to_verify = OpcodeHelper.get_jump_and_data(Opcode.JMP, -get_bytes_count(verify_while +
                                                                                         compare_item +
                                                                                         not_found), True)
        not_found.append(jmp_back_to_verify)

        jmp_to_error = OpcodeHelper.get_jump_and_data(Opcode.JMPLT, get_bytes_count(compare_item +
                                                                                    not_found), True)
        verify_while[-1] = jmp_to_error

        not_inside_sequence = [             # send error message saying that substring not found
            (Opcode.PUSHDATA1, Integer(len(message)).to_byte_array(signed=True, min_length=1) + message),
            (Opcode.THROW, b''),
        ]

        jmp_to_return_index = OpcodeHelper.get_jump_and_data(Opcode.JMPIF, get_bytes_count(not_found +
                                                                                           not_inside_sequence), True)
        compare_item[-1] = jmp_to_return_index

        return_index = [                    # removes all values in the stack but the index
            (Opcode.NIP, b''),
            (Opcode.NIP, b''),
            (Opcode.NIP, b''),
            (Opcode.NIP, b''),
        ]

        return (
            verify_and_fix_end +
            verify_and_fix_start +
            change_stack_order +
            verify_while +
            compare_item +
            not_found +
            not_inside_sequence +
            return_index
        )



from typing import Dict, Optional

from boa3.model.builtin.builtinproperty import IBuiltinProperty
from boa3.model.builtin.interop.contractgethashmethod import ContractGetHashMethod
from boa3.model.variable import Variable


class GetStdLibScriptHashMethod(ContractGetHashMethod):
    def __init__(self):
        from boa3.constants import STD_LIB_SCRIPT
        from boa3.model.type.collection.sequence.uint160type import UInt160Type
        identifier = '-get_std_lib_contract'
        args: Dict[str, Variable] = {}
        super().__init__(STD_LIB_SCRIPT, identifier, args, return_type=UInt160Type.build())

    @property
    def _args_on_stack(self) -> int:
        return len(self.args)

    @property
    def _body(self) -> Optional[str]:
        return None


class StdLibProperty(IBuiltinProperty):
    def __init__(self):
        identifier = 'StdLib'
        getter = GetStdLibScriptHashMethod()
        super().__init__(identifier, getter)


StdLibContract = StdLibProperty()



__all__ = ['GetDesignatedByRoleMethod',
           'RoleType'
           ]

from boa3.model.builtin.interop.role.getdesignatedbyrolemethod import GetDesignatedByRoleMethod
from boa3.model.builtin.interop.role.roletype import RoleType



__all__ = ['BooleanNot',
           'LogicNot',
           'Negative',
           'Positive'
           ]

from boa3.model.operation.unary.booleannot import BooleanNot
from boa3.model.operation.unary.logicnot import LogicNot
from boa3.model.operation.unary.negative import Negative
from boa3.model.operation.unary.positive import Positive
from boa3.model.operation.unary.unaryoperation import UnaryOperation



from typing import Dict, List, Tuple

from boa3.model.builtin.interop.stdlib import ItoaMethod
from boa3.model.builtin.method.strmethod import StrMethod
from boa3.model.variable import Variable
from boa3.neo.vm.opcode.Opcode import Opcode


class StrIntMethod(StrMethod):

    def __init__(self):
        from boa3.model.type.type import Type
        args: Dict[str, Variable] = {
            'object': Variable(Type.int),
        }

        super().__init__(args)

    @property
    def _opcode(self) -> List[Tuple[Opcode, bytes]]:
        return ItoaMethod(internal_call_args=1).opcode



from enum import IntEnum, IntFlag


class WitnessScope(IntFlag):
    """
    Determine the rules for a smart contract :func:`CheckWitness()` sys call.
    """
    #: No Contract was witnessed. Only sign the transaction.
    NONE = 0x0
    #: Allow the witness if the current calling script hash equals the entry script hash into the virtual machine.
    #: Using this prevents passing :func:`CheckWitness()` in a smart contract called via another smart contract.
    CALLED_BY_ENTRY = 0x01
    #: Allow the witness if called from a smart contract that is whitelisted in the signer
    #: :attr:`~neo3.network.payloads.verification.Signer.allowed_contracts` attribute.
    CUSTOM_CONTRACTS = 0x10
    #: Allow the witness if any public key is in the signer
    #: :attr:`~neo3.network.payloads.verification.Signer.allowed_groups` attribute is whitelisted in the contracts
    #: manifest.groups array.
    CUSTOM_GROUPS = 0x20
    #: Allow the witness if the specified :attr:`~neo3.network.payloads.verification.Signer.rules` are satisfied
    WITNESS_RULES = 0x40
    #: Allow the witness in all context. Equal to NEO 2.x's default behaviour.
    GLOBAL = 0x80

    def to_csharp_name(self) -> str:
        """
        Internal helper to match C# convention
        """
        if self == 0:
            return "None"
        flags = []
        if self.CALLED_BY_ENTRY in self:
            flags.append("CalledByEntry")
        if self.CUSTOM_CONTRACTS in self:
            flags.append("CustomContracts")
        if self.CUSTOM_GROUPS in self:
            flags.append("CustomGroups")
        if self.WITNESS_RULES in self:
            flags.append("WitnessRules")
        if self.GLOBAL in self:
            flags.append("Global")
        return ", ".join(flags)

    @classmethod
    def from_chsarp_name(cls, csharp_name):
        """
        Internal helper to parse from C# convention
        """
        c = cls(cls.NONE)
        if "CalledByEntry" in csharp_name:
            c |= c.CALLED_BY_ENTRY
        if "CustomContracts" in csharp_name:
            c |= c.CUSTOM_CONTRACTS
        if "CustomGroups" in csharp_name:
            c |= c.CUSTOM_GROUPS
        if "WitnessRules" in csharp_name:
            c |= c.WITNESS_RULES
        if "Global" in csharp_name:
            c |= c.GLOBAL
        return c


class WitnessRuleAction(IntEnum):
    DENY = 0
    ALLOW = 1


class WitnessConditionType(IntEnum):
    BOOLEAN = 0x0
    NOT = 0x01
    AND = 0x2
    OR = 0x03
    SCRIPT_HASH = 0x18
    GROUP = 0x19
    CALLED_BY_ENTRY = 0x20
    CALLED_BY_CONTRACT = 0x28
    CALLED_BY_GROUP = 0x29

    def to_csharp_string(self) -> str:
        if self == WitnessConditionType.SCRIPT_HASH:
            return "ScriptHash"
        elif self == WitnessConditionType.CALLED_BY_ENTRY:
            return "CalledByEntry"
        elif self == WitnessConditionType.CALLED_BY_CONTRACT:
            return "CalledByContract"
        elif self == WitnessConditionType.CALLED_BY_GROUP:
            return "CalledByGroup"
        else:
            return self.name.title()



import abc
from typing import List, Tuple

from boa3 import constants
from boa3.model.type.classes import classtype
from boa3.neo.vm.opcode.Opcode import Opcode
from boa3.neo.vm.type.AbiType import AbiType
from boa3.neo.vm.type.StackItem import StackItemType


class PythonClass(classtype.ClassType, abc.ABC):

    def __init__(self, identifier: str,
                 instance_variables: dict = None,
                 instance_methods: dict = None,
                 properties: dict = None,
                 class_variables: dict = None,
                 class_methods: dict = None,
                 static_methods: dict = None):

        self._instance_methods = instance_methods if isinstance(instance_methods, dict) else None
        self._instance_variables = instance_variables if isinstance(instance_variables, dict) else None
        self._properties = properties if isinstance(properties, dict) else None
        self._class_variables = class_variables if isinstance(class_variables, dict) else None
        self._class_methods = class_methods if isinstance(class_methods, dict) else None
        self._static_methods = static_methods if isinstance(static_methods, dict) else None

        is_init_defined = isinstance(instance_methods, dict) and constants.INIT_METHOD_ID in instance_methods
        self._is_init_set = is_init_defined
        self._constructor = (instance_methods[constants.INIT_METHOD_ID]
                             if is_init_defined
                             else None)

        super().__init__(identifier)

    def _init_class_symbols(self):
        """
        Overwrite this method to set variables and methods from this type.
        Always call super()._init_class_symbols in the beginning
        Used to avoid circular imports between the init classes
        """
        # TODO: May be removed when class inheritance is implemented
        if not isinstance(self._instance_methods, dict):
            self._instance_methods = {}
        if not isinstance(self._instance_variables, dict):
            self._instance_variables = {}
        if not isinstance(self._properties, dict):
            self._properties = {}
        if not isinstance(self._class_variables, dict):
            self._class_variables = {}
        if not isinstance(self._class_methods, dict):
            self._class_methods = {}
        if not isinstance(self._static_methods, dict):
            self._static_methods = {}

    @property
    def class_variables(self):
        if not isinstance(self._class_variables, dict):
            self._init_class_symbols()
        return self._class_variables.copy()

    @property
    def instance_variables(self):
        if not isinstance(self._instance_variables, dict):
            self._init_class_symbols()
        return self._instance_variables.copy()

    @property
    def properties(self):
        if not isinstance(self._properties, dict):
            self._init_class_symbols()
        return self._properties.copy()

    @property
    def static_methods(self):
        if not isinstance(self._static_methods, dict):
            self._init_class_symbols()
        return self._static_methods.copy()

    @property
    def class_methods(self):
        if not isinstance(self._class_methods, dict):
            self._init_class_symbols()
        return self._class_methods.copy()

    @property
    def instance_methods(self):
        if not isinstance(self._instance_methods, dict):
            self._init_class_symbols()
        return self._instance_methods.copy()

    def constructor_method(self):
        if not isinstance(self._instance_variables, dict):
            self._init_class_symbols()
        if not self._is_init_set:
            self._constructor = (self._instance_methods[constants.INIT_METHOD_ID]
                                 if constants.INIT_METHOD_ID in self._instance_methods
                                 else None)
            self._is_init_set = True
        return self._constructor

    @property
    def abi_type(self) -> AbiType:
        return super().abi_type

    @property
    def stack_item(self) -> StackItemType:
        """
        Get the Neo VM stack item type representation for this type

        :return: the stack item type of this type. Any by default.
        """
        return super().stack_item

    def is_instance_opcodes(self) -> List[Tuple[Opcode, bytes]]:
        """
        Get the list of opcodes to check if an value is of this type

        :return: A list of opcodes to check a value type
        """
        return [(Opcode.ISTYPE, self.stack_item)]



from typing import List, Optional

from boa3.model.builtin.method.builtinmethod import IBuiltinMethod
from boa3.model.type.type import Type
from boa3.neo3.contracts import CallFlags
from boa3.neo3.contracts.nef import MethodToken
from boa3.neo3.core import types


class MethodTokenCollection:
    def __init__(self):
        self._method_tokens: List[MethodToken] = []
        self._called_builtins: List[IBuiltinMethod] = []

    def append(self, contract_method: IBuiltinMethod, call_flag: CallFlags = CallFlags.ALL) -> Optional[int]:
        method_token_id = self._try_get_index(contract_method, call_flag)
        if method_token_id is None and hasattr(contract_method, 'contract_script_hash'):
            method_token = MethodToken(hash=types.UInt160(contract_method.contract_script_hash),
                                       method=contract_method.external_name,
                                       parameters_count=(contract_method.internal_call_args
                                                         if hasattr(contract_method, 'internal_call_args')
                                                         else len(contract_method.args)),
                                       has_return_value=contract_method.return_type is not Type.none,
                                       call_flags=call_flag)

            if contract_method not in self._called_builtins:
                self._called_builtins.append(contract_method)

            if hasattr(contract_method, 'method_name') and len(contract_method.method_name) == 0:
                # it's a contract method with a different interface, so it shouldn't have its method token included
                return None

            method_token_id = len(self._method_tokens)
            if hasattr(contract_method, '_method_token_id'):
                contract_method.reset()
            self._method_tokens.append(method_token)

        return method_token_id

    def clear(self):
        # reset the opcodes to ensure the correct output when calling consecutive compilations
        for method in self._called_builtins:
            method.reset()

        self._called_builtins.clear()
        return self._method_tokens.clear()

    def _try_get_index(self, contract_method: IBuiltinMethod, call_flag: CallFlags = CallFlags.ALL) -> Optional[int]:
        if not hasattr(contract_method, 'contract_script_hash'):
            return None

        parameters_count = (contract_method.internal_call_args
                            if hasattr(contract_method, 'internal_call_args')
                            else len(contract_method.args))

        method_token_index = next((index for index, token in enumerate(self._method_tokens)
                                   if (token.hash.to_array() == contract_method.contract_script_hash
                                       and token.method == contract_method.external_name
                                       and token.parameters_count == parameters_count
                                       and token.has_return_value == (contract_method.return_type is not Type.none)
                                       and token.call_flags == call_flag)), None)
        return method_token_index

    def to_list(self) -> List[MethodToken]:
        return self._method_tokens.copy()



from typing import Dict

from boa3.model.builtin.interop.blockchain.vmstatetype import VMStateType
from boa3.model.builtin.interop.nativecontract import LedgerMethod
from boa3.model.variable import Variable


class GetTransactionVMStateMethod(LedgerMethod):

    def __init__(self, vm_state_type: VMStateType):
        from boa3.model.type.collection.sequence.uint256type import UInt256Type

        identifier = 'get_transaction_vm_state'
        syscall = 'getTransactionVMState'
        args: Dict[str, Variable] = {'hash_': Variable(UInt256Type.build())}
        super().__init__(identifier, syscall, args, return_type=vm_state_type)



import ast
from typing import Dict, List, Optional, Tuple

from boa3.model.builtin.method.builtinmethod import IBuiltinMethod
from boa3.model.variable import Variable
from boa3.neo.vm.opcode.Opcode import Opcode


class SumMethod(IBuiltinMethod):

    def __init__(self):
        from boa3.model.type.type import Type
        identifier = 'sum'
        args: Dict[str, Variable] = {'__iterable': Variable(Type.sequence.build_collection(Type.int)),
                                     '__start': Variable(Type.int)}

        start_default = ast.parse("{0}".format(Type.int.default_value)
                                  ).body[0].value
        super().__init__(identifier, args, defaults=[start_default], return_type=Type.int)

    @property
    def _opcode(self) -> List[Tuple[Opcode, bytes]]:
        from boa3.neo.vm.type.Integer import Integer
        return [
            (Opcode.PUSH0, b''),    # index = 0
            (Opcode.DUP, b''),
            (Opcode.PUSH2, b''),    # len(iterable)
            (Opcode.PICK, b''),
            (Opcode.SIZE, b''),
            (Opcode.GE, b''),       # index > len(iterable)
            (Opcode.JMPIF, Integer(12).to_byte_array(signed=True, min_length=1)),
            (Opcode.REVERSE3, b''),
            (Opcode.OVER, b''),
            (Opcode.PUSH3, b''),    # x = iterable[index]
            (Opcode.PICK, b''),
            (Opcode.PICKITEM, b''),
            (Opcode.ADD, b''),      # result += x
            (Opcode.REVERSE3, b''),
            (Opcode.INC, b''),      # index += 1
            (Opcode.JMP, Integer(-15).to_byte_array(signed=True, min_length=1)),
            (Opcode.DROP, b''),
            (Opcode.DROP, b''),
        ]

    @property
    def _args_on_stack(self) -> int:
        return len(self.args)

    @property
    def _body(self) -> Optional[str]:
        return None



from boa3.constants import BYTEORDER


class Integer(int):
    def to_byte_array(self, signed: bool = False, min_length: int = 0) -> bytes:
        """
        Converts an integer value to a immutable byte array

        :param signed: Determines whether two's complement is used to represent the integer.
        :param min_length: Minimum output length.
        :return: Return an array of bytes representing an integer.
        """

        bits_per_byte: int = 8
        aux: int = bits_per_byte - 1
        if self < 0:
            signed = True
        if self > 0 and signed:
            aux += 1  # signed numbers uses an additional bit to represent the signal

        byte_length: int = ((self.bit_length() + aux) // bits_per_byte)
        length = max(min_length, byte_length)

        try:
            return int.to_bytes(self, length, BYTEORDER, signed=signed)
        except OverflowError:
            return int.to_bytes(self, 1 + length, BYTEORDER, signed=signed)

    @classmethod
    def from_bytes(cls, bts: bytes, signed: bool = False, byte_size: int = 0) -> int:
        if byte_size > 0:
            bts = bts[:byte_size]
        return int.from_bytes(bts, BYTEORDER, signed=signed)



import json
import logging
from typing import Any, Dict, List, Optional, Tuple

from boa3 import constants
from boa3.analyser.analyser import Analyser
from boa3.compiler.compileroutput import CompilerOutput
from boa3.model.event import Event
from boa3.model.imports.importsymbol import BuiltinImport, Import
from boa3.model.method import Method
from boa3.model.symbol import ISymbol
from boa3.model.type.classes.classtype import ClassType
from boa3.model.variable import Variable
from boa3.neo import to_hex_str
from boa3.neo.contracts.neffile import NefFile


class FileGenerator:
    """
    This class is responsible for generating the files.
    """

    def __init__(self, compiler_result: CompilerOutput, analyser: Analyser, entry_file: str):
        import os
        self._metadata = analyser.metadata
        self._symbols: Dict[str, ISymbol] = analyser.symbol_table.copy()

        self._entry_file = entry_file
        self._entry_file_full_path = analyser.path.replace(os.sep, constants.PATH_SEPARATOR)

        self._files: List[str] = [self._entry_file_full_path]
        self._nef: NefFile = NefFile(compiler_result.bytecode,
                                     source=self._metadata.source,
                                     method_tokens=compiler_result.method_tokens)

        self.__all_imports: List[Import] = None
        self._all_methods: Dict[str, Method] = None
        self._all_static_vars: Dict[str, Variable] = None

    @property
    def _public_methods(self) -> Dict[str, Method]:
        """
        Gets a sublist of the symbols containing all public methods

        :return: a dictionary that maps each public method with its identifier
        """
        return {name: method for name, method in self._methods.items() if method.is_public}

    @property
    def _static_variables(self) -> Dict[str, Variable]:
        """
        Gets a sublist of the symbols containing all global variables

        :return: a dictionary that maps each global variable with its identifier
        """
        if self._all_static_vars is not None:
            return self._all_static_vars
        else:
            variables: Dict[Tuple[str, str], Variable] = {}
            imported_symbols: Dict[str, Import] = {}

            for name, symbol in self._symbols.items():
                if isinstance(symbol, Variable) and not symbol.is_reassigned:
                    variables[(self._entry_file, name)] = symbol
                elif isinstance(symbol, Import):
                    imported_symbols[symbol.origin] = symbol

            imported_to_map, imports_unique_ids = self._get_imports_unique_ids(imported_symbols,
                                                                               False,
                                                                               list(variables.values())
                                                                               )

            if imports_unique_ids[-1] != self._entry_file:
                # update entry file to be a unique name
                unique_id = imports_unique_ids[-1]
                variables = {(unique_id, method_name): method for (module_id, method_name), method in variables.items()}

            # include all user created methods in the list, even the methods that aren't imported in the entry file
            for index in range(len(imported_to_map)):
                module_full_path, module_import = list(imported_to_map.items())[index]
                module_id = imports_unique_ids[index]

                for name, symbol in module_import.all_symbols.items():
                    need_to_map = (isinstance(symbol, Variable)
                                   and not symbol.is_reassigned
                                   and symbol not in [variables for variables in variables.values()])
                    if need_to_map:
                        variables[(module_id, name)] = symbol

            if len(imports_unique_ids) > 1:
                self._all_static_vars = {f'{unique_id}{constants.VARIABLE_NAME_SEPARATOR}{var_id}': var
                                         for (unique_id, var_id), var in variables.items()}
            else:
                self._all_static_vars = {var_id: var
                                         for (unique_id, var_id), var in variables.items()}

            return self._all_static_vars

    @property
    def _methods(self) -> Dict[str, Method]:
        """
        Gets a sublist of the symbols containing all user methods

        :return: a dictionary that maps each method with its identifier
        """
        from boa3.model.builtin.method import IBuiltinMethod
        return {name: method for name, method in self._symbols.items()
                if ((isinstance(method, Method) and method.defined_by_entry)
                    or (isinstance(method, IBuiltinMethod) and method.is_public)
                    )}

    @property
    def _methods_with_imports(self) -> Dict[Tuple[str, str], Method]:
        if self._all_methods is not None:
            return self._all_methods

        from boa3.model.builtin.decorator.builtindecorator import IBuiltinCallable

        methods: Dict[Tuple[str, str], Method] = {}
        imported_symbols: Dict[str, Import] = {}

        for name, symbol in self._symbols.items():
            if symbol.defined_by_entry and isinstance(symbol, Method) and not isinstance(symbol, IBuiltinCallable):
                methods[(self._entry_file, name)] = symbol
            elif isinstance(symbol, Import):
                imported_symbols[symbol.origin] = symbol

        imported_to_map, imports_unique_ids = self._get_imports_unique_ids(imported_symbols,
                                                                           True,
                                                                           list(methods.values())
                                                                           )

        if imports_unique_ids[-1] != self._entry_file:
            # update entry file to be a unique name
            unique_id = imports_unique_ids[-1]
            methods = {(unique_id, method_name): method for (module_id, method_name), method in methods.items()}

        # include all user created methods in the list, even the methods that aren't imported in the entry file
        for index in range(len(imported_to_map)):
            module_full_path, module_import = list(imported_to_map.items())[index]
            module_id = imports_unique_ids[index]

            for name, symbol in module_import.all_symbols.items():
                if (isinstance(symbol, Method)
                        and not isinstance(symbol, IBuiltinCallable)
                        and symbol not in [method for method in methods.values()]):
                    methods[(module_id, name)] = symbol

        self._all_methods = methods
        return methods

    @property
    def _events(self) -> Dict[str, Event]:
        """
        Gets a sublist of the symbols containing all user events

        :return: a dictionary that maps each event with its identifier
        """
        events = set()
        for imported in self._all_imports:
            events.update([event for event in imported.all_symbols.values() if isinstance(event, Event)])
        events.update([event for event in self._symbols.values() if isinstance(event, Event)])

        return {event.name: event for event in events}

    @property
    def _all_imports(self) -> List[Import]:
        if self.__all_imports is not None:
            return self.__all_imports

        all_imports = [imported for imported in self._symbols.values()
                       if (isinstance(imported, Import)
                           and not isinstance(imported, BuiltinImport))]
        index = 0
        while index < len(all_imports):
            imported = all_imports[index]
            for inner in imported.all_symbols.values():
                if (isinstance(inner, Import)
                        and not isinstance(inner, BuiltinImport)
                        and inner not in all_imports):
                    all_imports.append(inner)
            index += 1

        self.__all_imports = list(reversed(all_imports))  # first positions are the most inner imports
        # using for instead a generator to keep the result determined
        for import_ in all_imports:
            if import_.origin not in self._files:
                self._files.append(import_.origin)
        return self.__all_imports

    # region NEF

    @property
    def _nef_hash(self) -> str:
        """
        Gets the string representation of the hash of the nef file

        :return: the hex string representation of the hash
        """
        return to_hex_str(self._nef.script_hash)

    def generate_nef_file(self) -> bytes:
        """
        Generates the .nef file

        :return: the resulting nef file as a byte array
        """
        return self._nef.serialize()

    # endregion

    # region Manifest

    def generate_manifest_file(self) -> bytes:
        """
        Generates the .manifest metadata file

        :return: the resulting manifest as a byte array
        """
        data: Dict[str, Any] = self._get_manifest_info()
        json_data: str = json.dumps(data, indent=4)
        return bytes(json_data, constants.ENCODING)

    def _get_manifest_info(self) -> Dict[str, Any]:
        """
        Gets the manifest information in a dictionary format

        :return: a dictionary with the manifest information
        """
        # TODO: fill the information of the manifest
        return {
            "name": self._get_name(),
            "groups": self._get_groups(),
            "abi": self._get_abi_info(),
            "permissions": self._get_permissions(),
            "trusts": self._metadata.trusts,
            "features": {},
            "supportedstandards": self._metadata.supported_standards,
            "extra": self._get_extras()
        }

    def _get_name(self) -> str:
        """
        Gets the name of the contract, if name wasn't specified it will be the file name.

        :return: the contract name
        """
        return self._metadata.name if self._metadata.name else self._entry_file

    def _get_permissions(self) -> List[Dict[str, Any]]:
        """
        Gets the permission information in a dictionary format, if _metadata._permissions is empty, then consider it
        with the import wildcard inside it.

        :return: a dictionary with the permission information
        """
        return self._metadata.permissions if self._metadata.permissions else [{"contract": constants.IMPORT_WILDCARD,
                                                                               "methods": constants.IMPORT_WILDCARD}]

    def _get_groups(self) -> List[Dict[str, Any]]:
        """
        Gets the group information in a dictionary format.

        :return: a dictionary with the groups information
        """
        return self._metadata.groups

    def _get_abi_info(self) -> Dict[str, Any]:
        """
        Gets the abi information in a dictionary format

        :return: a dictionary with the abi information
        """
        return {
            "methods": self._get_abi_methods(),
            "events": self._get_abi_events()
        }

    def _get_abi_methods(self) -> List[Dict[str, Any]]:
        """
        Gets the abi methods in a dictionary format

        :return: a dictionary with the abi methods
        """
        methods = []
        for method_id, method in self._public_methods.items():
            methods.append(self._construct_abi_method(method_id, method))
        return methods

    def _construct_abi_method(self, method_id: str, method: Method) -> Dict[str, Any]:
        from boa3.compiler.codegenerator.vmcodemapping import VMCodeMapping

        abi_method_name = method.external_name if isinstance(method.external_name, str) else method_id
        logging.info(f"'{abi_method_name}' method included in the manifest")

        return {
            "name": abi_method_name,
            "offset": (VMCodeMapping.instance().get_start_address(method.start_bytecode)
                       if method.start_bytecode is not None else 0),
            "parameters": [
                {
                    "name": arg_id,
                    "type": arg.type.abi_type
                } for arg_id, arg in method.args.items()
            ],
            "returntype": method.type.abi_type,
            "safe": method.is_safe
        }

    def _get_abi_events(self) -> List[Dict[str, Any]]:
        """
        Gets the abi events in a dictionary format

        :return: a dictionary with the abi events
        """
        return [
            {
                "name": name,
                "parameters": [
                    {
                        "name": arg_id,
                        "type": arg.type.abi_type
                    } for arg_id, arg in event.args_to_generate.items()
                ],
            } for name, event in self._events.items() if event.is_called
        ]

    def _get_extras(self) -> Optional[Dict[str, Any]]:
        """
        Gets the abi information in a dictionary format

        :return: a dictionary with the abi information
        """
        extras = {}
        for key, value in self._metadata.extras.items():
            try:
                json.dumps(value)  # include only json like extras
                extras[key] = value
            except BaseException:
                continue

        return extras if len(extras) > 0 else None

    # endregion

    # region Debug Info

    def generate_nefdbgnfo_file(self) -> bytes:
        """
        Generates a debug map for NEO debugger

        :return: the resulting map as a byte array
        """
        data: Dict[str, Any] = self._get_debug_info()
        json_data: str = json.dumps(data, indent=4)
        return bytes(json_data, constants.ENCODING)

    def _get_debug_info(self) -> Dict[str, Any]:
        """
        Gets the debug information in a dictionary format

        :return: a dictionary with the debug information
        """
        return {
            "hash": self._nef_hash,
            "documents": self._files,
            "static-variables": self._get_debug_static_variables(),
            "methods": self._get_debug_methods(),
            "events": self._get_debug_events()
        }

    def _get_debug_methods(self) -> List[Dict[str, Any]]:
        """
        Gets the methods' debug information in a dictionary format

        :return: a dictionary with the methods' debug information
        """
        return [
            self._get_method_debug_info(module_id, method_id, method)
            for (module_id, method_id), method in self._methods_with_imports.items()
            if method.is_compiled
        ]

    def _get_method_debug_info(self, module_id: str, method_id: str, method: Method) -> Dict[str, Any]:
        from boa3.compiler.codegenerator.vmcodemapping import VMCodeMapping
        from boa3.neo.vm.type.AbiType import AbiType
        from boa3.model.type.itype import IType
        return {
            "id": str(id(method)),
            "name": '{0},{1}'.format(module_id, method_id),
            "range": '{0}-{1}'.format(method.start_address, method.end_address),
            "params": [
                '{0},{1}'.format(name, var.type.abi_type) for name, var in method.args.items()
            ],
            "return": method.return_type.abi_type,
            "variables": [
                '{0},{1}'.format(name, var.type.abi_type if isinstance(var.type, IType) else AbiType.Any)
                for name, var in method.locals.items()
            ],
            "sequence-points": [
                '{0}[{1}]{2}:{3}-{4}:{5}'.format(VMCodeMapping.instance().get_start_address(instruction.code),
                                                 self._get_method_origin_index(method),
                                                 instruction.start_line, instruction.start_col,
                                                 instruction.end_line, instruction.end_col)
                for instruction in method.debug_map()
            ]
        }

    def _get_method_origin_index(self, method: Method) -> int:
        imported_files: List[Import] = [imported for imported in self._symbols.values()
                                        if (isinstance(imported, Import)
                                            and not isinstance(imported, BuiltinImport)
                                            and imported.origin is not None)]
        imported = None
        for file in imported_files:
            if method in file.all_symbols.values():
                imported = file
                break

        if imported is None:
            return 0
        else:
            if imported.origin not in self._files:
                self._files.append(imported.origin)
            return self._files.index(imported.origin)

    def _get_debug_events(self) -> List[Dict[str, Any]]:
        """
        Gets the events' debug information in a dictionary format

        :return: a dictionary with the event's debug information
        """
        return [
            {
                "id": str(id(event)),
                "name": ',{0}'.format(event_id),  # TODO: include module name
                "params": [
                    '{0},{1}'.format(name, var.type.abi_type) for name, var in event.args.items()
                ]
            } for event_id, event in self._events.items()
        ]

    def _get_debug_static_variables(self) -> List[str]:
        """
        Gets the static variables' debug information in a dictionary format

        :return: a dictionary with the event's debug information
        """
        from boa3.model.type.itype import IType
        from boa3.neo.vm.type.AbiType import AbiType

        static_variables = []

        for name, var in self._static_variables.items():
            var_unique_name = self._get_static_var_unique_name(name)
            var_type = var.type.abi_type if isinstance(var.type, IType) else AbiType.Any

            values = [var_unique_name, var_type.name]

            var_slot_index = self._get_static_var_slot_index(name)
            if isinstance(var_slot_index, int):
                values.append(str(var_slot_index))

            static_variables.append(
                constants.VARIABLE_NAME_SEPARATOR.join(values)
            )

        return static_variables

    # endregion

    def _get_static_var_unique_name(self, variable_id) -> str:
        imported_symbols: Dict[str, Import] = {}

        for name, symbol in self._symbols.items():
            if isinstance(symbol, Import):
                imported_symbols[symbol.origin] = symbol

        imported_to_map, imports_unique_ids = self._get_imports_unique_ids(imported_symbols,
                                                                           False,
                                                                           list(self._static_variables.values())
                                                                           )
        split_name = variable_id.split(constants.VARIABLE_NAME_SEPARATOR)
        if len(split_name) > 1:
            variable_original_id = split_name[-1]
            imported_id = constants.VARIABLE_NAME_SEPARATOR.join(split_name[:-1])
        else:
            variable_original_id = variable_id
            imported_id = None

        if isinstance(imported_id, str) and imported_id in imports_unique_ids:
            return '{0}.{1}'.format(imported_id, variable_original_id)

        if len(imported_to_map) <= 1:
            return variable_original_id

        for index, imported in enumerate(imported_symbols.values()):
            if isinstance(imported, Import) and variable_original_id in imported.all_symbols:
                if len(split_name) <= 1 or str(imported.ast.__hash__()) == split_name[0]:
                    return '{0}.{1}'.format(imports_unique_ids[index], variable_original_id)

        return '{0}.{1}'.format(imports_unique_ids[-1], variable_original_id)

    def _get_static_var_slot_index(self, variable_id) -> Optional[int]:
        module_globals = list(self._static_variables.keys())
        if variable_id in module_globals:
            return module_globals.index(variable_id)
        return None

    def _get_imports_unique_ids(self, imported_symbols: Dict[str, Import],
                                importing_methods: bool,
                                inner_imported_symbols: List[ISymbol] = None) -> Tuple[Dict[str, Import], List[str]]:
        if not isinstance(imported_symbols, dict):
            return {}, []
        if not isinstance(inner_imported_symbols, list):
            inner_imported_symbols = []

        from boa3.model.builtin.builtincallable import IBuiltinCallable

        # must map all imports, including inner imports
        index = 0
        while index < len(imported_symbols):
            module_origin, imported = list(imported_symbols.items())[index]
            for name, inner_imported in imported.all_symbols.items():
                if isinstance(inner_imported, Import) and inner_imported.origin not in inner_imported_symbols:
                    imported_symbols[inner_imported.origin] = inner_imported
            index += 1

        # map the modules that have user modules not imported by the entry file
        imported_to_map: Dict[str, Import] = {}
        for name, imported in imported_symbols.items():
            if isinstance(imported, BuiltinImport):
                need_to_map = False
            else:
                need_to_map = any(not importing_methods  # is importing variables or is a method but not builtin
                                  or (((isinstance(symbol, Method) and not isinstance(symbol, IBuiltinCallable))
                                       or (isinstance(symbol, ClassType) and len(symbol.symbols) > 0)
                                       )
                                      and symbol not in inner_imported_symbols)
                                  for name, symbol in imported.all_symbols.items())
            if need_to_map:
                filtered_name = name.replace('.py', '').replace('/__init__', '')
                imported_to_map[filtered_name] = imported_symbols[name]
                if name not in self._files:
                    self._files.append(name)

        # change the full path names to unique small names
        imports_paths = list(imported_to_map)
        imports_paths.append(self._entry_file_full_path.replace('.py', '').replace('/__init__', ''))

        imports_unique_ids = []
        imports_duplicated_ids = []

        for index in range(len(imports_paths)):
            name = imports_paths[index]
            split_name = name.split(constants.PATH_SEPARATOR)
            index = -1
            short_name = split_name[index]
            while short_name in imports_duplicated_ids:
                index -= 1
                short_name = constants.ATTRIBUTE_NAME_SEPARATOR.join(split_name[index:])

            if short_name in imports_unique_ids:
                index_of_duplicated = imports_unique_ids.index(short_name)

                dup_split_name = imports_paths[index_of_duplicated].split(constants.PATH_SEPARATOR)
                dup_new_id = short_name
                dup_index = -len(imports_unique_ids[index].split(constants.ATTRIBUTE_NAME_SEPARATOR))

                while dup_new_id == short_name:
                    imports_duplicated_ids.append(short_name)
                    index -= 1
                    dup_index -= 1
                    short_name = constants.ATTRIBUTE_NAME_SEPARATOR.join(split_name[index:])
                    dup_new_id = constants.ATTRIBUTE_NAME_SEPARATOR.join(dup_split_name[dup_index:])

                imports_unique_ids[index_of_duplicated] = dup_new_id

            imports_unique_ids.append(short_name)

        return imported_to_map, imports_unique_ids



from __future__ import annotations

import enum


class WitnessScope(enum.IntEnum):
    # Indicates that no contract was witnessed. Only sign the transaction.
    _None = 0

    # Indicates that the calling contract must be the entry contract. The witness/permission/signature
    # given on first invocation will automatically expire if entering deeper internal
    # invokes. This can be the default safe choice for native NEO/GAS (previously used
    # on Neo 2 as "attach" mode).
    CalledByEntry = 1

    # Custom hash for contract-specific.
    CustomContracts = 16

    # Custom pubkey for group members.
    CustomGroups = 32

    # Indicates that the current context must satisfy the specified rules.
    WitnessRules = 64

    # This allows the witness in all contexts (default Neo2 behavior).
    Global = 128

    def neo_name(self) -> str:
        if self is WitnessScope._None:
            return 'None'
        else:
            return self.name

    @staticmethod
    def get_from_neo_name(neo_name: str) -> WitnessScope:
        if neo_name == 'None':
            return WitnessScope._None
        elif neo_name != WitnessScope._None.name:
            return WitnessScope[neo_name]



from typing import Any, List

from boa3.builtin.compile_time import public
from boa3_test.test_sc.import_test.FromImportTyping import empty_list


@public
def Main() -> List[Any]:
    return empty_list



from typing import Any, List

from boa3.builtin.compile_time import public
from boa3_test.test_sc.import_test.FromImportTyping import EmptyList

a = EmptyList()


@public
def Main() -> List[Any]:
    return a



from __future__ import annotations

from typing import Any, Dict

from boa3.model.builtin.interop.nativecontract import LedgerContract
from boa3.model.builtin.native.inativecontractclass import INativeContractClass
from boa3.model.method import Method


class LedgerClass(INativeContractClass):
    """
    A class used to represent Ledger native contract
    """

    def __init__(self):
        super().__init__('Ledger', LedgerContract)

    @property
    def class_methods(self) -> Dict[str, Method]:
        # avoid recursive import
        from boa3.model.builtin.interop.interop import Interop

        if len(self._class_methods) == 0:
            self._class_methods = {
                'get_block': Interop.GetBlock,
                'get_current_index': Interop.CurrentIndex.getter,
                'get_transaction': Interop.GetTransaction,
                'get_transaction_from_block': Interop.GetTransactionFromBlock,
                'get_transaction_height': Interop.GetTransactionHeight,
                'get_transaction_signers': Interop.GetTransactionSigners,
                'get_transaction_vm_state': Interop.GetTransactionVMState
            }
        return super().class_methods

    @classmethod
    def build(cls, value: Any = None) -> LedgerClass:
        if value is None or cls._is_type_of(value):
            return _Ledger

    @classmethod
    def _is_type_of(cls, value: Any):
        return isinstance(value, LedgerClass)


_Ledger = LedgerClass()



from boa3.builtin.compile_time import public


@public
def main(bytes_value: bytes, subbytes_value: bytes, start: int, end: int) -> bool:
    return bytes_value.startswith(subbytes_value, start, end)



from __future__ import annotations

from typing import Any, Dict, List

from boa3 import constants
from boa3.neo import from_hex_str, to_hex_str
from boa3.neo3.core.types import UInt160
from boa3_test.tests.test_classes.witnessscope import WitnessScope


class Signer:
    def __init__(self, account: UInt160, scopes: WitnessScope = WitnessScope.CalledByEntry):
        if scopes == WitnessScope.WitnessRules:
            raise NotImplementedError
        self._account: UInt160 = account
        self._scopes: WitnessScope = scopes
        self._allowed_contracts: List[UInt160] = []
        self._allowed_groups: List[bytes] = []
        # TODO: implement allowed witness rules

    @property
    def account(self) -> UInt160:
        return self._account

    @property
    def scopes(self) -> WitnessScope:
        return self._scopes

    def to_json(self) -> Dict[str, Any]:
        return {
            'account': str(self._account),
            'scopes': self._scopes.neo_name(),
            'allowedcontracts': [str(contract_hash) for contract_hash in self._allowed_contracts],
            'allowedgroups': [to_hex_str(group_pubkey) for group_pubkey in self._allowed_groups],
            'rules': []
        }

    @classmethod
    def from_json(cls, json: Dict[str, Any]) -> Signer:
        account_hex = json['account']
        account = UInt160(from_hex_str(account_hex))
        scopes = WitnessScope.get_from_neo_name(json['scopes']) if 'scopes' in json else WitnessScope.CalledByEntry

        signer = cls(account, scopes)
        signer._allowed_contracts = ([UInt160(from_hex_str(contract_hex)) for contract_hex in json['allowedcontracts']]
                                     if 'allowedcontracts' in json else [])
        signer._allowed_groups = ([from_hex_str(group_hex) for group_hex in json['allowedgroups']]
                                  if 'allowedgroups' in json else [])
        # TODO: implement allowed witness rules
        return signer

    def __str__(self) -> str:
        return self._account.__str__()

    def __eq__(self, other) -> bool:
        return isinstance(other, Signer) and self._account == other._account

    def add_scope(self, scope: WitnessScope):
        if scope == WitnessScope.WitnessRules:
            raise NotImplementedError
        if scope > self._scopes:
            self._scopes = scope

    def add_permitted_contracts(self, contract_hashes: list):
        for contract_hash in contract_hashes:
            if isinstance(contract_hash, bytes):
                script_hash = UInt160(contract_hash)
            elif isinstance(contract_hash, str):
                script_hash = UInt160.from_string(contract_hash)
            else:
                script_hash = contract_hash

            if not isinstance(script_hash, UInt160):
                raise ValueError(f"Incorrect data type for allowed contract: {type(script_hash).__name__}")
            if script_hash not in self._allowed_contracts:
                self._allowed_contracts.append(script_hash)

    def add_permitted_groups(self, groups_pubkeys: List[bytes]):
        for group_pubkey in groups_pubkeys:
            if isinstance(group_pubkey, str):
                group = from_hex_str(group_pubkey)
            else:
                group = group_pubkey

            if not isinstance(group, bytes):
                raise ValueError(f"Incorrect data type for allowed group: {type(group).__name__}")
            elif len(group) != constants.SIZE_OF_ECPOINT:
                raise ValueError(f"Incorrect data for allowed group")
            else:
                group = bytes(group)

            if group not in self._allowed_groups:
                self._allowed_groups.append(group)

    def add_permitted_rules(self, custom_rules: list):
        raise NotImplementedError

    def set_permissions(self, permissions: list):
        if self._scopes == WitnessScope.CustomContracts:
            self.add_permitted_contracts(permissions)
        elif self._scopes == WitnessScope.CustomGroups:
            self.add_permitted_groups(permissions)
        elif self._scopes == WitnessScope.WitnessRules:
            self.add_permitted_rules(permissions)



from typing import Any, Dict, List, Optional, Tuple

from boa3.model.builtin.method.builtinmethod import IBuiltinMethod
from boa3.model.expression import IExpression
from boa3.model.type.collection.sequence.sequencetype import SequenceType
from boa3.model.type.collection.sequence.tupletype import TupleType
from boa3.model.type.itype import IType
from boa3.model.variable import Variable
from boa3.neo.vm.opcode.Opcode import Opcode


class PrintMethod(IBuiltinMethod):

    def __init__(self, arg_value: Optional[IType] = None):
        from boa3.model.type.type import Type
        identifier = 'print'
        if not isinstance(arg_value, IType):
            arg_value = Type.str

        args: Dict[str, Variable] = {}
        vararg = ('values', Variable(arg_value))
        super().__init__(identifier, args, return_type=Type.none, vararg=vararg)

        self._print_value_opcodes = None

    @property
    def _arg_values(self) -> Variable:
        return self._vararg[1]

    @property
    def identifier(self) -> str:
        from boa3.model.type.type import Type
        if self._arg_values.type is Type.str:
            return self._identifier
        return '-{0}_from_{1}'.format(self._identifier, self._arg_values.type._identifier)

    def validate_parameters(self, *params: IExpression) -> bool:
        if len(params) != 1:
            return False
        if not isinstance(params[0], IExpression):
            return False
        return isinstance(params[0].type, SequenceType)

    @property
    def _opcode(self) -> List[Tuple[Opcode, bytes]]:
        from boa3.compiler.codegenerator import get_bytes_count
        from boa3.model.builtin.interop.interop import Interop
        from boa3.neo.vm.type.Integer import Integer

        check_if_arg_is_empty = [
            (Opcode.DUP, b''),
            (Opcode.SIZE, b''),
            (Opcode.PUSH0, b''),
        ]
        copy_list_and_reverse = [
            (Opcode.UNPACK, b''),  # copy - it must not change the original
            (Opcode.PACK, b''),
            (Opcode.DUP, b''),
            (Opcode.REVERSEITEMS, b'')
        ]

        complete_loop = (
            [
                (Opcode.DUP, b''),
                (Opcode.POPITEM, b''),
            ] + self.print_value_opcodes
            + Interop.Log.opcode
            + check_if_arg_is_empty
        )
        complete_loop.append((Opcode.JMPNE, Integer(-get_bytes_count(complete_loop)).to_byte_array(signed=True)))

        return (check_if_arg_is_empty +
                [
                    (Opcode.JMPEQ, Integer(get_bytes_count(copy_list_and_reverse)
                                           + 1  # the size of the JMPEQ arg
                                           + get_bytes_count(complete_loop)).to_byte_array(signed=True)),
                ] +
                copy_list_and_reverse +
                complete_loop +
                [
                    (Opcode.DROP, b'')
                ])

    @property
    def print_value_opcodes(self) -> List[Tuple[Opcode, bytes]]:
        if self._print_value_opcodes is None:
            self._print_value_opcodes = []

        return self._print_value_opcodes

    @property
    def _args_on_stack(self) -> int:
        return len(self.args)

    @property
    def _body(self) -> Optional[str]:
        return None

    def build(self, value: Any) -> IBuiltinMethod:
        if isinstance(value, list) and len(value) == 1:
            value = value[0]
        if isinstance(value, TupleType):
            value = value.value_type
        if type(value) == type(self._arg_values.type):
            return self

        from boa3.model.builtin.method.printbytestringmethod import PrintByteStringMethod
        from boa3.model.type.type import Type
        from boa3.model.type.primitive.bytestringtype import ByteStringType

        if Type.bool.is_type_of(value):
            from boa3.model.builtin.method.printboolmethod import PrintBoolMethod
            return PrintBoolMethod()

        elif Type.int.is_type_of(value):
            from boa3.model.builtin.method.printintmethod import PrintIntMethod
            return PrintIntMethod()

        elif ByteStringType.build().is_type_of(value):
            return PrintByteStringMethod(value)

        elif Type.sequence.is_type_of(value):
            from boa3.model.builtin.method.printsequencemethod import PrintSequenceMethod
            return PrintSequenceMethod(value)

        return PrintByteStringMethod(value)



from boa3.boa3 import Boa3
from boa3_test.tests.boa_test import BoaTest
from boa3_test.tests.test_classes.testengine import TestEngine


class TestTemplate(BoaTest):
    default_folder: str = 'examples'

    def test_hello_world_compile(self):
        path = self.get_contract_path('hello_world.py')
        Boa3.compile(path)

    def test_hello_world_main(self):
        path = self.get_contract_path('hello_world.py')
        engine = TestEngine()
        result = self.run_smart_contract(engine, path, 'Main')
        self.assertIsVoid(result)

        storage_value = engine.storage_get(b'hello', path)
        self.assertIsNotNone(storage_value)
        self.assertEqual(b'world', storage_value)



from boa3.boa3 import Boa3
from boa3.exception import CompilerError
from boa3.neo.vm.opcode.Opcode import Opcode
from boa3.neo.vm.type.Integer import Integer
from boa3.neo.vm.type.StackItem import StackItemType
from boa3.neo.vm.type.String import String
from boa3_test.tests.boa_test import BoaTest
from boa3_test.tests.test_classes.TestExecutionException import TestExecutionException
from boa3_test.tests.test_classes.testengine import TestEngine


class TestAssert(BoaTest):
    default_folder: str = 'test_sc/assert_test'

    def test_assert_unary_boolean_operation(self):
        expected_output = (
            Opcode.INITSLOT     # function signature
            + b'\x00'
            + b'\x02'
            + Opcode.LDARG0     # assert not a
            + Opcode.NOT
            + Opcode.ASSERT
            + Opcode.LDARG1     # return b
            + Opcode.RET
        )

        path = self.get_contract_path('AssertUnaryOperation.py')
        output = Boa3.compile(path)
        self.assertEqual(expected_output, output)

        engine = TestEngine()
        result = self.run_smart_contract(engine, path, 'Main', False, 10)
        self.assertEqual(10, result)

        with self.assertRaisesRegex(TestExecutionException, self.ASSERT_RESULTED_FALSE_MSG):
            self.run_smart_contract(engine, path, 'Main', True, 20)

    def test_assert_binary_boolean_operation(self):
        expected_output = (
            Opcode.INITSLOT     # function signature
            + b'\x00'
            + b'\x02'
            + Opcode.LDARG0     # assert a != b
            + Opcode.LDARG1
            + Opcode.NUMNOTEQUAL
            + Opcode.ASSERT
            + Opcode.LDARG0     # return a
            + Opcode.RET
        )

        path = self.get_contract_path('AssertBinaryOperation.py')
        output = Boa3.compile(path)
        self.assertEqual(expected_output, output)

        engine = TestEngine()
        result = self.run_smart_contract(engine, path, 'Main', 10, 20)
        self.assertEqual(10, result)

        with self.assertRaisesRegex(TestExecutionException, self.ASSERT_RESULTED_FALSE_MSG):
            self.run_smart_contract(engine, path, 'Main', 20, 20)

    def test_assert_with_message(self):
        path = self.get_contract_path('AssertWithMessage.py')
        engine = TestEngine()

        result = self.run_smart_contract(engine, path, 'Main', 10)
        self.assertEqual(10, result)

        with self.assertRaisesRegex(TestExecutionException, self.ASSERT_RESULTED_FALSE_MSG):
            self.run_smart_contract(engine, path, 'Main', -10)

    def test_assert_with_bytes_message(self):
        path = self.get_contract_path('AssertWithBytesMessage.py')
        engine = TestEngine()

        result = self.run_smart_contract(engine, path, 'Main', 10)
        self.assertEqual(10, result)

        with self.assertRaisesRegex(TestExecutionException, self.ASSERT_RESULTED_FALSE_MSG):
            self.run_smart_contract(engine, path, 'Main', -10)

    def test_assert_with_int_message(self):
        path = self.get_contract_path('AssertWithIntMessage.py')
        self.assertCompilerLogs(CompilerError.MismatchedTypes, path)

    def test_assert_with_bool_message(self):
        path = self.get_contract_path('AssertWithBoolMessage.py')
        self.assertCompilerLogs(CompilerError.MismatchedTypes, path)

    def test_assert_with_list_message(self):
        path = self.get_contract_path('AssertWithListMessage.py')
        self.assertCompilerLogs(CompilerError.MismatchedTypes, path)

    def test_assert_with_str_var_message(self):
        path = self.get_contract_path('AssertWithStrVarMessage.py')
        engine = TestEngine()

        result = self.run_smart_contract(engine, path, 'Main', 10)
        self.assertEqual(10, result)

        with self.assertRaisesRegex(TestExecutionException, self.ASSERT_RESULTED_FALSE_MSG):
            self.run_smart_contract(engine, path, 'Main', -10)

    def test_assert_with_str_function_message(self):
        path = self.get_contract_path('AssertWithStrFunctionMessage.py')
        engine = TestEngine()

        result = self.run_smart_contract(engine, path, 'Main', 10)
        self.assertEqual(10, result)

        with self.assertRaisesRegex(TestExecutionException, self.ASSERT_RESULTED_FALSE_MSG):
            self.run_smart_contract(engine, path, 'Main', -10)

    def test_assert_int(self):
        expected_output = (
            Opcode.INITSLOT     # function signature
            + b'\x00'
            + b'\x01'
            + Opcode.LDARG0     # assert a
            + Opcode.ASSERT
            + Opcode.LDARG0     # return a
            + Opcode.RET
        )

        path = self.get_contract_path('AssertInt.py')
        output = Boa3.compile(path)
        self.assertEqual(expected_output, output)

        engine = TestEngine()
        result = self.run_smart_contract(engine, path, 'Main', 10)
        self.assertEqual(10, result)
        result = self.run_smart_contract(engine, path, 'Main', -10)
        self.assertEqual(-10, result)

        with self.assertRaisesRegex(TestExecutionException, self.ASSERT_RESULTED_FALSE_MSG):
            self.run_smart_contract(engine, path, 'Main', 0)

    def test_assert_str(self):
        expected_output = (
            Opcode.INITSLOT     # function signature
            + b'\x00'
            + b'\x01'
            + Opcode.LDARG0     # assert a
            + Opcode.ASSERT
            + Opcode.LDARG0     # return a
            + Opcode.RET
        )

        path = self.get_contract_path('AssertStr.py')
        output = Boa3.compile(path)
        self.assertEqual(expected_output, output)

        engine = TestEngine()
        result = self.run_smart_contract(engine, path, 'Main', 'unittest')
        self.assertEqual('unittest', result)

        with self.assertRaisesRegex(TestExecutionException, self.ASSERT_RESULTED_FALSE_MSG):
            self.run_smart_contract(engine, path, 'Main', '')

    def test_assert_bytes(self):
        expected_output = (
            Opcode.INITSLOT     # function signature
            + b'\x00'
            + b'\x01'
            + Opcode.LDARG0     # assert a
            + Opcode.ASSERT
            + Opcode.LDARG0     # return a
            + Opcode.RET
        )

        path = self.get_contract_path('AssertBytes.py')
        output = Boa3.compile(path)
        self.assertEqual(expected_output, output)

        engine = TestEngine()
        result = self.run_smart_contract(engine, path, 'Main', b'unittest')
        self.assertEqual(b'unittest', String(result).to_bytes())

        with self.assertRaisesRegex(TestExecutionException, self.ASSERT_RESULTED_FALSE_MSG):
            self.run_smart_contract(engine, path, 'Main', b'')

    def test_assert_list(self):
        expected_output = (
            Opcode.INITSLOT     # function signature
            + b'\x00'
            + b'\x01'
            + Opcode.LDARG0     # assert a
            + Opcode.SIZE
            + Opcode.ASSERT
            + Opcode.LDARG0     # return len(a)
            + Opcode.SIZE
            + Opcode.RET
        )

        path = self.get_contract_path('AssertList.py')
        output = Boa3.compile(path)
        self.assertEqual(expected_output, output)

        engine = TestEngine()
        result = self.run_smart_contract(engine, path, 'Main', [1, 2, 3])
        self.assertEqual(3, result)

        with self.assertRaisesRegex(TestExecutionException, self.ASSERT_RESULTED_FALSE_MSG):
            self.run_smart_contract(engine, path, 'Main', [])

    def test_assert_dict(self):
        expected_output = (
            Opcode.INITSLOT     # function signature
            + b'\x00'
            + b'\x01'
            + Opcode.LDARG0     # assert a
            + Opcode.SIZE
            + Opcode.ASSERT
            + Opcode.LDARG0     # return len(a)
            + Opcode.SIZE
            + Opcode.RET
        )

        path = self.get_contract_path('AssertDict.py')
        output = Boa3.compile(path)
        self.assertEqual(expected_output, output)

        engine = TestEngine()
        result = self.run_smart_contract(engine, path, 'Main', {1: 2, 2: 5})
        self.assertEqual(2, result)

        with self.assertRaisesRegex(TestExecutionException, self.ASSERT_RESULTED_FALSE_MSG):
            self.run_smart_contract(engine, path, 'Main', {})

    def test_assert_any(self):
        expected_output = (
            Opcode.INITSLOT     # function signature
            + b'\x00'
            + b'\x01'
            + Opcode.LDARG0     # assert a
            + Opcode.DUP
            + Opcode.ISTYPE + StackItemType.Array
            + Opcode.JMPIF + Integer(12).to_byte_array(signed=True, min_length=1)
            + Opcode.DUP
            + Opcode.ISTYPE + StackItemType.Map
            + Opcode.JMPIF + Integer(7).to_byte_array(signed=True, min_length=1)
            + Opcode.DUP
            + Opcode.ISTYPE + StackItemType.Struct
            + Opcode.JMPIFNOT + Integer(3).to_byte_array(signed=True, min_length=1)
            + Opcode.SIZE
            + Opcode.ASSERT
            + Opcode.RET
        )

        path = self.get_contract_path('AssertAny.py')
        output = Boa3.compile(path)
        self.assertEqual(expected_output, output)

        engine = TestEngine()
        result = self.run_smart_contract(engine, path, 'Main', True)
        self.assertIsVoid(result)

    def test_boa2_throw_test(self):
        path = self.get_contract_path('ThrowBoa2Test.py')
        engine = TestEngine()

        result = self.run_smart_contract(engine, path, 'main', 1)
        self.assertEqual(True, result)

        with self.assertRaisesRegex(TestExecutionException, self.ASSERT_RESULTED_FALSE_MSG):
            self.run_smart_contract(engine, path, 'main', 4)



import ast
import logging
import os
from abc import ABC
from inspect import isclass
from typing import Any, Dict, List, Optional, Sequence, Union

from boa3.exception.CompilerError import CompilerError, InternalError, UnresolvedReference
from boa3.exception.CompilerWarning import CompilerWarning
from boa3.model.attribute import Attribute
from boa3.model.expression import IExpression
from boa3.model.identifiedsymbol import IdentifiedSymbol
from boa3.model.operation.operation import IOperation
from boa3.model.symbol import ISymbol
from boa3.model.type.annotation.metatype import MetaType
from boa3.model.type.classes.classtype import ClassType
from boa3.model.type.type import IType, Type
from boa3.model.type.typeutils import TypeUtils


class IAstAnalyser(ABC, ast.NodeVisitor):
    """
    An interface for the analysers that walk the Python abstract syntax tree

    :ivar errors: a list that contains all the errors raised by the compiler. Empty by default.
    :ivar warnings: a list that contains all the warnings found by the compiler. Empty by default.
    """

    def __init__(self, ast_tree: ast.AST, filename: str = None, root_folder: str = None, log: bool = False):
        self.errors: List[CompilerError] = []
        self.warnings: List[CompilerWarning] = []

        self.filename: Optional[str] = filename
        if not isinstance(root_folder, str) or not os.path.isdir(root_folder):
            root_folder = (os.path.dirname(os.path.abspath(filename))
                           if isinstance(filename, str) and os.path.isfile(filename)
                           else os.path.abspath(os.path.curdir))
        self.root_folder: str = root_folder
        self._log: bool = log

        self._tree: ast.AST = ast_tree
        self.symbols: Dict[str, ISymbol] = {}

    @property
    def has_errors(self) -> bool:
        return len(self.errors) > 0

    def _log_error(self, error: CompilerError):
        if error.filepath is None:
            error.filepath = self.filename
        if not any(err == error for err in self.errors):
            # don't include duplicated errors
            self.errors.append(error)
            if self._log:
                logging.error(error)

    def _log_warning(self, warning: CompilerWarning):
        if warning.filepath is None:
            warning.filepath = self.filename
        if not any(warn == warning for warn in self.warnings):
            # don't include duplicated warnings
            self.warnings.append(warning)
            if self._log:
                logging.warning(warning)

    def _log_info(self, info_message: str, log_filename: bool = True):
        if self._log:
            if log_filename and self.filename:
                formatted_message = f'{info_message} <{self.filename}>'
            else:
                formatted_message = info_message

            logging.info(formatted_message)

    def visit(self, node: ast.AST) -> Any:
        try:
            return super().visit(node)
        except CompilerError as error:
            self._log_error(error)
        except CompilerWarning as warning:
            self._log_warning(warning)
        except BaseException as exception:
            if hasattr(node, 'lineno'):
                self._log_error(
                    InternalError(line=node.lineno,
                                  col=node.col_offset,
                                  raised_exception=exception)
                )

    def get_type(self, value: Any, use_metatype: bool = False) -> IType:
        """
        Returns the type of the given value.

        :param value: value to get the type
        :param use_metatype: whether it should return `Type.type` if the value is an IType implementation.
        :return: Returns the :class:`IType` of the the type of the value. `Type.none` by default.
        """
        # visits if it is a node
        if isinstance(value, ast.AST):
            fun_rtype_id: Any = ast.NodeVisitor.visit(self, value)
            if isinstance(fun_rtype_id, ast.Name):
                fun_rtype_id = fun_rtype_id.id

            if isinstance(fun_rtype_id, str) and not isinstance(value, ast.Str):
                value = self.get_symbol(fun_rtype_id, origin_node=value)
                if isinstance(value, IType) and not isinstance(value, MetaType):
                    value = TypeUtils.type.build(value) if use_metatype else value
            else:
                value = fun_rtype_id

        if isinstance(value, Attribute):
            if ((isinstance(value.attr_symbol, IExpression) and isinstance(value.attr_symbol.type, ClassType))
                    or (isinstance(value.attr_symbol, IType))):
                value = value.attr_symbol
            elif isinstance(value.type, IType):
                value = value.type

        if isinstance(value, IType):
            final_type = value
        elif isinstance(value, IExpression):
            final_type = value.type
        elif isinstance(value, IOperation):
            final_type = value.result
        else:
            final_type = Type.get_type(value)

        if isinstance(final_type, MetaType) and not use_metatype:
            return final_type.meta_type
        else:
            return final_type

    def get_symbol(self, symbol_id: str,
                   is_internal: bool = False,
                   check_raw_id: bool = False,
                   origin_node: ast.AST = None) -> Optional[ISymbol]:
        """
        Tries to get the symbol by its id name

        :param symbol_id: the id name of the symbol
        :return: the symbol if found. None otherwise.
        :rtype: ISymbol or None
        """
        if symbol_id in self.symbols:
            # the symbol exists in the global scope
            return self.symbols[symbol_id]

        if check_raw_id:
            found_symbol = self._search_by_raw_id(symbol_id, list(self.symbols.values()))
            if found_symbol is not None:
                # the symbol exists in the global scope, but with an alias different from the original name
                return found_symbol

        if is_internal:
            from boa3.model import imports
            found_symbol = imports.builtin.get_internal_symbol(symbol_id)
            if isinstance(found_symbol, ISymbol):
                return found_symbol

        # the symbol may be a built in. If not, returns None
        from boa3.model.builtin.builtin import Builtin
        found_symbol = Builtin.get_symbol(symbol_id)

        if found_symbol is None and isinstance(symbol_id, str) and self.is_exception(symbol_id):
            found_symbol = Builtin.Exception.return_type

        if origin_node is not None and found_symbol is None:
            self._log_error(
                UnresolvedReference(
                    line=origin_node.lineno,
                    col=origin_node.col_offset,
                    symbol_id=symbol_id
                )
            )

        return found_symbol

    def _search_by_raw_id(self, symbol_id: str, symbols: Sequence[ISymbol]) -> Optional[ISymbol]:
        for symbol in symbols:
            if isinstance(symbol, IdentifiedSymbol) and symbol.identifier == symbol_id:
                return symbol

        return None

    def is_exception(self, symbol_id: str) -> bool:
        global_symbols = globals()
        if symbol_id in global_symbols or symbol_id in global_symbols['__builtins__']:
            symbol = (global_symbols[symbol_id]
                      if symbol_id in global_symbols
                      else global_symbols['__builtins__'][symbol_id])
            if isclass(symbol) and issubclass(symbol, BaseException):
                return True
        return False

    def is_implemented_class_type(self, symbol) -> bool:
        if not isinstance(symbol, ClassType):
            return False

        from boa3.model.type.classes.pythonclass import PythonClass
        from boa3.model.builtin.interop.interopinterfacetype import InteropInterfaceType

        return not isinstance(symbol, PythonClass) or isinstance(symbol, InteropInterfaceType)

    def parse_to_node(self, expression: str, origin: ast.AST = None) -> Union[ast.AST, Sequence[ast.AST]]:
        """
        Parses an expression to an ast.

        :param expression: string expression to be parsed
        :param origin: an existing ast. If not None, the parsed node will have the same location of origin.
        :return: the parsed node
        :rtype: ast.AST or Sequence[ast.AST]
        """
        node = ast.parse(expression, filename='<{0}>'.format(type(self).__name__))
        if origin is not None:
            self.update_line_and_col(node, origin)

        # get the expression instead of the default root node
        if hasattr(node, 'body'):
            node = node.body
        elif hasattr(node, 'argtypes'):
            node = node.argtypes

        if isinstance(node, list) and len(node) == 1:
            # the parsed node has a list of expression and only one expression is found
            result = node[0]
        else:
            result = node

        if isinstance(result, ast.Expr):
            # an expr node encapsulates another node in its value field.
            result = result.value
        return result

    def update_line_and_col(self, target: ast.AST, origin: ast.AST):
        """
        Updates the position of a node and its child nodes

        :param target: the node that will have its position updated
        :param origin: the node with the desired position
        """
        ast.copy_location(target, origin)
        for field, value in ast.iter_fields(target):
            if isinstance(value, list):
                for item in value:
                    if isinstance(item, ast.AST):
                        self.update_line_and_col(item, origin)
            elif isinstance(value, ast.AST):
                self.update_line_and_col(value, origin)
        ast.fix_missing_locations(target)

    def clone(self, node: ast.AST) -> ast.AST:
        """
        Clones an AST node

        :param node: node to be cloned
        :return:
        """
        clone: ast.AST = node.__class__()
        clone._attributes = node._attributes
        clone._fields = node._fields

        for attr in node.__dict__:
            clone.__setattr__(attr, node.__getattribute__(attr))

        return clone
